# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import logging as log
import os
import os.path as osp
from typing import Optional

from datumaro.components.annotation import AnnotationType, Label, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer, with_subset_dirs
from datumaro.components.media import Image
from datumaro.util.definitions import SUBSET_NAME_BLACKLIST
from datumaro.util.image import find_images


class ImagenetPath:
    IMAGE_DIR_NO_LABEL = "no_label"
    SEP_TOKEN = ":"


class ImagenetBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isdir(path):
            raise NotADirectoryError(errno.ENOTDIR, "Can't find dataset directory", path)

        super().__init__(subset=subset, ctx=ctx)

        self._categories = self._load_categories(path)
        self._items = list(self._load_items(path).values())

    def _load_categories(self, path):
        label_cat = LabelCategories()
        for dirname in sorted(os.listdir(path)):
            if not os.path.isdir(os.path.join(path, dirname)):
                raise NotADirectoryError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), path)
            if dirname != ImagenetPath.IMAGE_DIR_NO_LABEL:
                label_cat.add(dirname)
        return {AnnotationType.label: label_cat}

    def _load_items(self, path):
        items = {}

        # Images should be in root/label_dir/*.img and root/*.img is not allowed.
        # => max_depth=1, min_depth=1
        for image_path in find_images(path, recursive=True, max_depth=1, min_depth=1):
            label = osp.basename(osp.dirname(image_path))
            image_name = osp.splitext(osp.basename(image_path))[0]

            item_id = label + ImagenetPath.SEP_TOKEN + image_name
            item = items.get(item_id)
            try:
                if item is None:
                    item = DatasetItem(
                        id=item_id, subset=self._subset, media=Image.from_file(path=image_path)
                    )
                    items[item_id] = item
            except Exception as e:
                self._ctx.error_policy.report_item_error(e, item_id=(item_id, self._subset))
            annotations = item.annotations

            if label != ImagenetPath.IMAGE_DIR_NO_LABEL:
                try:
                    label = self._categories[AnnotationType.label].find(label)[0]
                    annotations.append(Label(label=label))
                except Exception as e:
                    self._ctx.error_policy.report_annotation_error(
                        e, item_id=(item_id, self._subset)
                    )

        return items


class ImagenetImporter(Importer):
    """TorchVision's ImageFolder style importer.
    For example, it imports the following directory structure.

    .. code-block:: text

        root
        ├── label_0
        │   ├── label_0_1.jpg
        │   └── label_0_2.jpg
        └── label_1
            └── label_1_1.jpg

    """

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        # Images must not be under a directory whose name is blacklisted.
        for dname in os.listdir(context.root_path):
            dpath = osp.join(context.root_path, dname)
            if osp.isdir(dpath) and dname.lower() in SUBSET_NAME_BLACKLIST:
                context.fail(
                    f"{dname} is found in {context.root_path}. "
                    "However, Images must not be under a directory whose name is blacklisted "
                    f"(SUBSET_NAME_BLACKLIST={SUBSET_NAME_BLACKLIST})."
                )

        return super().detect(context)

    @classmethod
    def find_sources(cls, path):
        if not osp.isdir(path):
            return []

        # Images should be in root/label_dir/*.img and root/*.img is not allowed.
        # => max_depth=1, min_depth=1
        for _ in find_images(path, recursive=True, max_depth=1, min_depth=1):
            return [{"url": path, "format": ImagenetBase.NAME}]

        return []


@with_subset_dirs
class ImagenetWithSubsetDirsImporter(ImagenetImporter):
    """TorchVision ImageFolder style importer.
    For example, it imports the following directory structure.

    .. code-block::

        root
        ├── train
        │   ├── label_0
        │   │   ├── label_0_1.jpg
        │   │   └── label_0_2.jpg
        │   └── label_1
        │       └── label_1_1.jpg
        ├── val
        │   ├── label_0
        │   │   ├── label_0_1.jpg
        │   │   └── label_0_2.jpg
        │   └── label_1
        │       └── label_1_1.jpg
        └── test
            ├── label_0
            │   ├── label_0_1.jpg
            │   └── label_0_2.jpg
            └── label_1
                └── label_1_1.jpg

    Then, it will have three subsets: train, val, and test and they have label_0 and label_1 labels.
    """


class ImagenetExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".jpg"
    USE_SUBSET_DIRS = False

    def _apply_impl(self):
        def _get_name(item: DatasetItem) -> str:
            id_parts = item.id.split(ImagenetPath.SEP_TOKEN)

            if len(id_parts) == 1:
                # e.g. item.id = my_img_1
                return item.id
            else:
                # e.g. item.id = label_1:my_img_1
                return "_".join(id_parts[1:])  # ":" is not allowed in windows

        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        if 1 < len(self._extractor.subsets()) and not self.USE_SUBSET_DIRS:
            log.warning(
                f"There are more than one subset in the dataset ({len(self._extractor.subsets())}). "
                "However, ImageNet format exports all dataset items into the same directory. "
                "Therefore, subset information will be lost. To prevent it, please use ImagenetWithSubsetDirsExporter. "
                'For example, dataset.export("<path/to/output>", format="imagenet_with_subset_dirs").'
            )

        root_dir = self._save_dir
        extractor = self._extractor
        labels = {}
        for item in self._extractor:
            file_name = _get_name(item)
            labels = set(p.label for p in item.annotations if p.type == AnnotationType.label)

            for label in labels:
                label_name = extractor.categories()[AnnotationType.label][label].name
                self._save_image(
                    item,
                    subdir=osp.join(root_dir, item.subset, label_name)
                    if self.USE_SUBSET_DIRS
                    else osp.join(root_dir, label_name),
                    name=file_name,
                )

            if not labels:
                self._save_image(
                    item,
                    subdir=osp.join(root_dir, item.subset, ImagenetPath.IMAGE_DIR_NO_LABEL)
                    if self.USE_SUBSET_DIRS
                    else osp.join(root_dir, ImagenetPath.IMAGE_DIR_NO_LABEL),
                    name=file_name,
                )


class ImagenetWithSubsetDirsExporter(ImagenetExporter):
    USE_SUBSET_DIRS = True



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Optional, Union

import numpy as np
from attr import attrib, attrs

from datumaro.components.abstracts import IMergerContext
from datumaro.components.abstracts.merger import IMatcherContext
from datumaro.components.annotation import Annotation
from datumaro.util.annotation_util import (
    OKS,
    approximate_line,
    bbox_iou,
    max_bbox,
    mean_bbox,
    segment_iou,
)

__all__ = [
    "match_segments_pair",
    "match_segments_more_than_pair",
    "AnnotationMatcher",
    "LabelMatcher",
    "ShapeMatcher",
    "BboxMatcher",
    "PolygonMatcher",
    "MaskMatcher",
    "PointsMatcher",
    "LineMatcher",
    "CaptionsMatcher",
    "Cuboid3dMatcher",
    "ImageAnnotationMatcher",
    "HashKeyMatcher",
    "FeatureVectorMatcher",
]


def match_segments_pair(
    a_segms,
    b_segms,
    distance=segment_iou,
    dist_thresh=1.0,
    label_matcher=lambda a, b: a.label == b.label,
):
    """Match segments and return pairs of the two matched segments"""

    assert callable(distance), distance
    assert callable(label_matcher), label_matcher

    a_segms.sort(key=lambda ann: 1 - ann.attributes.get("score", 1))
    b_segms.sort(key=lambda ann: 1 - ann.attributes.get("score", 1))

    # a_matches: indices of b_segms matched to a bboxes
    # b_matches: indices of a_segms matched to b bboxes
    a_matches = -np.ones(len(a_segms), dtype=int)
    b_matches = -np.ones(len(b_segms), dtype=int)

    distances = np.array([[distance(a, b) for b in b_segms] for a in a_segms])

    # matches: boxes we succeeded to match completely
    # mispred: boxes we succeeded to match, having label mismatch
    matches = []
    mispred = []

    # It needs len(a_segms) > 0 and len(b_segms) > 0
    if len(b_segms) > 0:
        for a_idx, a_segm in enumerate(a_segms):
            matched_b = -1
            max_dist = -1
            b_indices = np.argsort(
                [not label_matcher(a_segm, b_segm) for b_segm in b_segms], kind="stable"
            )  # prioritize those with same label, keep score order
            for b_idx in b_indices:
                if 0 <= b_matches[b_idx]:  # assign a_segm with max conf
                    continue
                d = distances[a_idx, b_idx]
                if d < dist_thresh or d <= max_dist:
                    continue
                max_dist = d
                matched_b = b_idx

            if matched_b < 0:
                continue
            a_matches[a_idx] = matched_b
            b_matches[matched_b] = a_idx

            b_segm = b_segms[matched_b]

            if label_matcher(a_segm, b_segm):
                matches.append((a_segm, b_segm))
            else:
                mispred.append((a_segm, b_segm))

    # *_umatched: boxes of (*) we failed to match
    a_unmatched = [a_segms[i] for i, m in enumerate(a_matches) if m < 0]
    b_unmatched = [b_segms[i] for i, m in enumerate(b_matches) if m < 0]

    return matches, mispred, a_unmatched, b_unmatched


def match_segments_more_than_pair(
    a_segms,
    b_segms,
    distance=segment_iou,
    dist_thresh=1.0,
    label_matcher=lambda a, b: a.label == b.label,
):
    """Match segments and return sets of the matched segments which can be more than two"""

    assert callable(distance), distance
    assert callable(label_matcher), label_matcher

    # a_matches: indices of b_segms matched to a bboxes
    # b_matches: indices of a_segms matched to b bboxes
    a_matches = -np.ones(len(a_segms), dtype=int)
    b_matches = -np.ones(len(b_segms), dtype=int)

    distances = np.array([[distance(a, b) for b in b_segms] for a in a_segms])

    # matches: boxes we succeeded to match completely
    # mispred: boxes we succeeded to match, having label mismatch
    matches = []
    mispred = []

    # It needs len(a_segms) > 0 and len(b_segms) > 0
    if len(b_segms) > 0:
        for a_idx, a_segm in enumerate(a_segms):
            b_indices = np.argsort(
                [not label_matcher(a_segm, b_segm) for b_segm in b_segms], kind="stable"
            )  # prioritize those with same label, keep score order
            for b_idx in b_indices:
                d = distances[a_idx, b_idx]
                if d < dist_thresh:
                    continue

                a_matches[a_idx] = b_idx
                b_matches[b_idx] = a_idx

                b_segm = b_segms[b_idx]

                if label_matcher(a_segm, b_segm):
                    matches.append((a_segm, b_segm))
                else:
                    mispred.append((a_segm, b_segm))

    # *_umatched: boxes of (*) we failed to match
    a_unmatched = [a_segms[i] for i, m in enumerate(a_matches) if m < 0]
    b_unmatched = [b_segms[i] for i, m in enumerate(b_matches) if m < 0]

    return matches, mispred, a_unmatched, b_unmatched


@attrs(kw_only=True)
class AnnotationMatcher:
    _context: Optional[Union[IMatcherContext, IMergerContext]] = attrib(default=None)

    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class LabelMatcher(AnnotationMatcher):
    def distance(self, a, b):
        a_label = self._context.get_any_label_name(a, a.label)
        b_label = self._context.get_any_label_name(b, b.label)
        return a_label == b_label

    def match_annotations(self, sources):
        return [sum(sources, [])]


@attrs(kw_only=True)
class ShapeMatcher(AnnotationMatcher):
    pairwise_dist = attrib(converter=float, default=0.9)
    cluster_dist = attrib(converter=float, default=-1.0)
    _match_segments = attrib(default=match_segments_pair)

    def match_annotations(self, sources: List[List[Annotation]]) -> List[List[Annotation]]:
        distance = self.distance
        label_matcher = self.label_matcher
        pairwise_dist = self.pairwise_dist
        cluster_dist = self.cluster_dist

        if cluster_dist < 0:
            cluster_dist = pairwise_dist

        id_segm = {id(a): (a, id(s)) for s in sources for a in s}

        def _is_close_enough(cluster, extra_id):
            # check if whole cluster IoU will not be broken
            # when this segment is added
            b = id_segm[extra_id][0]
            for a_id in cluster:
                a = id_segm[a_id][0]
                if distance(a, b) < cluster_dist:
                    return False
            return True

        def _has_same_source(cluster, extra_id):
            b = id_segm[extra_id][1]
            for a_id in cluster:
                a = id_segm[a_id][1]
                if a == b:
                    return True
            return False

        # match segments in sources, pairwise
        adjacent = {i: [] for i in id_segm}  # id(sgm) -> [id(adj_sgm1), ...]
        for a_idx, src_a in enumerate(sources):
            for src_b in sources[a_idx + 1 :]:
                matches, _, _, _ = self._match_segments(
                    src_a,
                    src_b,
                    dist_thresh=pairwise_dist,
                    distance=distance,
                    label_matcher=label_matcher,
                )
                for a, b in matches:
                    adjacent[id(a)].append(id(b))

        # join all segments into matching clusters
        clusters = []
        visited = set()
        for cluster_idx in adjacent:
            if cluster_idx in visited:
                continue

            cluster = set()
            to_visit = {cluster_idx}
            while to_visit:
                c = to_visit.pop()
                cluster.add(c)
                visited.add(c)

                for i in adjacent[c]:
                    if i in visited:
                        continue
                    if 0 < cluster_dist and not _is_close_enough(cluster, i):
                        continue
                    if _has_same_source(cluster, i):
                        continue

                    to_visit.add(i)

            clusters.append([id_segm[i][0] for i in cluster])

        return clusters

    def distance(self, a, b):
        return segment_iou(a, b)

    def label_matcher(self, a, b):
        a_label = self._context.get_any_label_name(a, a.label)
        b_label = self._context.get_any_label_name(b, b.label)
        return a_label == b_label


@attrs
class BboxMatcher(ShapeMatcher):
    pass


@attrs
class PolygonMatcher(ShapeMatcher):
    pass


@attrs
class MaskMatcher(ShapeMatcher):
    pass


@attrs(kw_only=True)
class PointsMatcher(ShapeMatcher):
    sigma: Optional[list] = attrib(default=None)
    instance_map = attrib(converter=dict)

    def distance(self, a, b):
        a_bbox = self.instance_map[id(a)][1]
        b_bbox = self.instance_map[id(b)][1]
        if bbox_iou(a_bbox, b_bbox) <= 0:
            return 0
        bbox = mean_bbox([a_bbox, b_bbox])
        return OKS(a, b, sigma=self.sigma, bbox=bbox)


@attrs
class LineMatcher(ShapeMatcher):
    def distance(self, a, b):
        # Compute inter-line area by using the Trapezoid formulae
        # https://en.wikipedia.org/wiki/Trapezoidal_rule
        # Normalize by common bbox and get the bbox fill ratio
        # Call this ratio the "distance"

        # The box area is an early-exit filter for non-intersected figures
        bbox = max_bbox([a, b])
        box_area = bbox[2] * bbox[3]
        if not box_area:
            return 1

        def _approx(line, segments):
            if len(line) // 2 != segments + 1:
                line = approximate_line(line, segments=segments)
            return np.reshape(line, (-1, 2))

        segments = max(len(a.points) // 2, len(b.points) // 2, 5) - 1

        a = _approx(a.points, segments)
        b = _approx(b.points, segments)
        dists = np.linalg.norm(a - b, axis=1)
        dists = dists[:-1] + dists[1:]
        a_steps = np.linalg.norm(a[1:] - a[:-1], axis=1)
        b_steps = np.linalg.norm(b[1:] - b[:-1], axis=1)

        # For the common bbox we can't use
        # - the AABB (axis-alinged bbox) of a point set
        # - the exterior of a point set
        # - the convex hull of a point set
        # because these soultions won't be correctly normalized.
        # The lines can have multiple self-intersections, which can give
        # the inter-line area more than internal area of the options above,
        # producing the value of the distance outside of the [0; 1] range.
        #
        # Instead, we can compute the upper boundary for the inter-line
        # area based on the maximum point distance and line length.
        max_area = np.max(dists) * max(np.sum(a_steps), np.sum(b_steps))

        area = np.dot(dists, a_steps + b_steps) * 0.5 * 0.5 / max(max_area, 1.0)

        return abs(1 - area)


@attrs
class CaptionsMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class Cuboid3dMatcher(ShapeMatcher):
    def distance(self, a, b):
        raise NotImplementedError()


@attrs
class ImageAnnotationMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class HashKeyMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class FeatureVectorMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class TabularMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import copy
import errno
import os
from collections.abc import Iterator
from typing import Dict, List, Optional, Tuple, TypeVar, Union, overload

import numpy as np
import pyarrow as pa

PathLike = TypeVar("PathLike", str, os.PathLike, None)


# TODO: consider replacing with `bisect` library
def _binary_search(arr, x, low=None, high=None, fn=None):
    low = 0 if low is None else low
    high = len(arr) - 1 if high is None else high
    if fn is None:

        def fn(arr, x):
            return arr[x]

    while low <= high:
        mid = (high + low) // 2
        val = fn(arr, mid)
        if val > x:
            high = mid - 1
        else:
            low = mid + 1
    return high


class ArrowDatasetIterator(Iterator):
    def __init__(self, dataset: "ArrowDataset"):
        self.dataset = dataset
        self.index = 0

    def __next__(self) -> Dict:
        if self.index >= len(self.dataset):
            raise StopIteration
        item = self.dataset[self.index]
        self.index += 1
        return item


class ArrowDataset:
    def __init__(
        self,
        files: Union[PathLike, List[PathLike]],
        keep_in_memory: bool = False,
    ):
        assert files
        if not isinstance(files, (list, tuple)):
            files = [files]

        self._files = files
        self._keep_in_memory = keep_in_memory

        self.unload_arrows()
        self.load_arrows()

    @property
    def table(self) -> pa.Table:
        self.load_arrows()
        return self.__table

    @property
    def column_names(self) -> List[str]:
        return self.table.column_names

    @property
    def num_columns(self) -> int:
        return self.table.num_columns

    @property
    def num_rows(self) -> int:
        return self.table.num_rows

    @property
    def shape(self) -> Tuple[int]:
        return self.table.shape

    @property
    def metadata(self) -> Optional[Dict[bytes, bytes]]:
        return self.table.schema.metadata

    def __deepcopy__(self, memo) -> "ArrowDataset":
        args = [self._files, self._keep_in_memory]
        copied_args = copy.deepcopy(args, memo)
        return self.__class__(*copied_args)

    def __len__(self) -> int:
        return self.table.num_rows

    def __iter__(self):
        return ArrowDatasetIterator(self)

    def unload_arrows(self) -> None:
        self.__table: pa.Table = None
        self.__batches: List[pa.RecordBatch] = []
        self.__offsets: List[int] = []

    def load_arrows(self) -> None:
        if self.__table is not None:
            return

        table = None
        for file in self._files:
            if not os.path.exists(file):
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file)

            if self._keep_in_memory:
                stream = pa.input_stream(file)
            else:
                stream = pa.memory_map(file)
            opened_stream = pa.ipc.open_stream(stream)
            table_ = opened_stream.read_all()
            if table:
                table = pa.concat_tables((table, table_))
            else:
                table = table_
        self._update_table(table)

    def _update_table(self, table: pa.Table) -> None:
        self.__table = table
        self.__batches = self.__table.to_batches()
        self.__offsets = np.cumsum([0] + [len(i) for i in self.__batches]).tolist()

    def flatten(self) -> "ArrowDataset":
        dataset = ArrowDataset(self._files, self._keep_in_memory)
        dataset._update_table(dataset.table.flatten())
        return dataset

    def get_batches(self, offset: int, length: int) -> List[pa.RecordBatch]:
        assert length > 0

        s_idx = _binary_search(self.__offsets, offset)
        e_idx = _binary_search(self.__offsets, offset + length - 1)

        batches = self.__batches[s_idx : e_idx + 1]
        if self.__offsets[s_idx + 1] >= offset + length:
            batches[0] = batches[0].slice(offset - self.__offsets[s_idx], length)
        else:
            batches[0] = batches[0].slice(offset - self.__offsets[s_idx])
            batches[-1] = batches[-1].slice(0, offset + length - self.__offsets[e_idx])

        return batches

    def slice(self, offset: int, length: int) -> List[Dict]:
        batches = self.get_batches(offset, length)
        items = pa.Table.from_batches(batches).to_pylist()
        return items

    @overload
    def __getitem__(self, key: str) -> "ArrowDataset":
        ...

    @overload
    def __getitem__(self, key: int) -> Dict:
        ...

    @overload
    def __getitem__(self, key: Union[range, slice]) -> List[Dict]:
        ...

    def __getitem__(
        self, key: Union[int, range, slice, str]
    ) -> Union[Dict, List[Dict], "ArrowDataset"]:
        if isinstance(key, str):
            if key not in self.column_names:
                raise KeyError(key)
            dataset = ArrowDataset(self._files, self._keep_in_memory)
            dataset._update_table(
                self.table.drop([name for name in self.column_names if name != key])
            )
            return dataset
        elif isinstance(key, slice):
            key = range(*key.indices(len(self)))

        if isinstance(key, range):
            if key.step == 1 and key.stop >= key.start:
                return self.slice(key.start, key.stop - key.start)
            return [
                self.slice(key_, 1)[0] if key_ >= 0 else self.slice(key_ + len(self), 1)[0]
                for key_ in key
            ]
        if isinstance(key, int):
            if key < 0:
                key += len(self)
            return self.slice(key, 1)[0]

        raise KeyError(key)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"num_rows={len(self)}, "
            f"columns={self.column_names}, "
            f"keep_in_memory={self._keep_in_memory}"
            ")"
        )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
import struct
from typing import List, Optional

import pyarrow as pa

from datumaro.components.dataset_base import SubsetBase
from datumaro.components.errors import MediaTypeError
from datumaro.components.importer import ImportContext
from datumaro.components.media import MediaType
from datumaro.components.merge import get_merger
from datumaro.plugins.data_formats.datumaro.base import JsonReader
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import DictMapper

from .arrow_dataset import ArrowDataset
from .mapper.dataset_item import DatasetItemMapper


class ArrowBase(SubsetBase):
    def __init__(
        self,
        path: str,
        additional_paths: Optional[List[str]] = None,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(subset=subset, ctx=ctx)

        self._paths = [path]
        if additional_paths:
            self._paths += additional_paths

        self._load()

    def _load(self):
        infos = []
        categories = []
        media_types = set()

        for path in self._paths:
            with pa.ipc.open_stream(path) as reader:
                schema = reader.schema

                _infos, _ = DictMapper.backward(schema.metadata.get(b"infos", b"\x00\x00\x00\x00"))
                infos.append(JsonReader._load_infos({"infos": _infos}))

                _categories, _ = DictMapper.backward(
                    schema.metadata.get(b"categories", b"\x00\x00\x00\x00")
                )
                categories.append(JsonReader._load_categories({"categories": _categories}))

                (media_type,) = struct.unpack(
                    "<I", schema.metadata.get(b"media_type", b"\x00\x00\x00\x00")
                )
                media_types.add(MediaType(media_type).media)

                dataset = ArrowDataset(path)
                dataset = dataset.flatten()

                for i in self._ctx.progress_reporter.iter(
                    range(len(dataset)), desc=f"Reading arrow from '{osp.basename(path)}'"
                ):
                    batches = dataset.get_batches(i, 1)
                    self._items.extend(DatasetItemMapper.backward_from_batches(batches))

        if len(media_types) > 1:
            raise MediaTypeError("Datasets have different media types")
        merger = get_merger("exact")
        self._infos = merger.merge_infos(infos)
        self._categories = merger.merge_categories(categories)
        self._media_type = list(media_types)[0]



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import math
import random
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple

import numpy as np
from sklearn.cluster import KMeans

import datumaro.plugins.ndr as ndr
from datumaro.components.algorithms.hash_key_inference.base import HashInference
from datumaro.components.algorithms.hash_key_inference.hashkey_util import (
    calculate_hamming,
    format_templates,
    select_uninferenced_dataset,
    templates,
)
from datumaro.components.annotation import HashKey, Label, LabelCategories
from datumaro.components.dataset import Dataset
from datumaro.components.dataset_base import DatasetItem


def match_num_item_for_cluster(ratio, dataset_len, cluster_num_item_list):
    total_num_selected_item = math.ceil(dataset_len * ratio)

    cluster_weights = np.array(cluster_num_item_list) / sum(cluster_num_item_list)
    norm_cluster_num_item_list = (cluster_weights * total_num_selected_item).astype(int)
    remaining_items = total_num_selected_item - sum(norm_cluster_num_item_list)

    if remaining_items > 0:
        zero_cluster_indexes = np.where(norm_cluster_num_item_list == 0)[0]
        add_clust_dist = np.sort(cluster_weights[zero_cluster_indexes])[::-1][:remaining_items]

        for dist in set(add_clust_dist):
            indices = np.where(cluster_weights == dist)[0]
            for index in indices:
                norm_cluster_num_item_list[index] += 1

    elif remaining_items < 0:
        diff_num_item_list = np.argsort(cluster_weights - norm_cluster_num_item_list)
        for diff_idx in diff_num_item_list[: abs(remaining_items)]:
            norm_cluster_num_item_list[diff_idx] -= 1

    return norm_cluster_num_item_list.tolist()


class PruneBase(ABC):
    @abstractmethod
    def base(
        self,
        ratio: float,
        num_centers: Optional[int],
        labels: Optional[List[int]],
        database_keys: Optional[np.ndarray],
        item_list: List[DatasetItem],
        source: Optional[Dataset],
    ) -> Tuple[List[DatasetItem], Optional[Dict]]:
        """It executes each method for pruning.

        Parameters:
            ratio: How much to remain dataset after pruning.
            num_centers: Number of centers for clustering.
            labels: Label of one annotation for each datasetitem.
            database_keys: Batch of the numpy formatted hash_key.
            item_list: List of datasetitem of dataset.
            source: Whole dataset.
        Returns:
            It returns a tuple of selected items and distance of each item and clusters.
        """
        raise NotImplementedError


class RandomSelect(PruneBase):
    """
    Select items randomly from the dataset.
    """

    def base(self, ratio, num_centers, labels, database_keys, item_list, source):
        random.seed(0)
        dataset_len = len(item_list)
        num_selected_item = math.ceil(dataset_len * ratio)
        random_indices = random.sample(range(dataset_len), num_selected_item)
        selected_items = [item_list[idx] for idx in random_indices]
        return selected_items, None


class Centroid(PruneBase):
    """
    Select items through clustering with centers targeting the desired number.
    """

    def base(self, ratio, num_centers, labels, database_keys, item_list, source):
        num_selected_centers = math.ceil(len(item_list) * ratio)
        kmeans = KMeans(n_clusters=num_selected_centers, random_state=0)
        clusters = kmeans.fit_predict(database_keys)
        cluster_centers = kmeans.cluster_centers_
        cluster_ids = np.unique(clusters)

        selected_items = []
        dist_tuples = []
        for cluster_id in cluster_ids:
            cluster_center = cluster_centers[cluster_id]
            cluster_items_idx = np.where(clusters == cluster_id)[0]
            num_selected_items = 1
            cluster_items = database_keys[cluster_items_idx,]
            dist = calculate_hamming(cluster_center, cluster_items)
            ind = np.argsort(dist)
            item_idx_list = cluster_items_idx[ind]
            for i, idx in enumerate(item_idx_list[:num_selected_items]):
                selected_items.append(item_list[idx])
                dist_tuples.append(
                    (cluster_id, item_list[idx].id, item_list[idx].subset, dist[ind][i])
                )
        return selected_items, dist_tuples


class ClusteredRandom(PruneBase):
    """
    Select items through clustering and choose randomly within each cluster.
    """

    def base(self, ratio, num_centers, labels, database_keys, item_list, source):
        kmeans = KMeans(n_clusters=num_centers, random_state=0)
        clusters = kmeans.fit_predict(database_keys)
        cluster_ids, cluster_num_item_list = np.unique(clusters, return_counts=True)

        norm_cluster_num_item_list = match_num_item_for_cluster(
            ratio, len(database_keys), cluster_num_item_list
        )

        selected_items = []
        random.seed(0)
        for i, cluster_id in enumerate(cluster_ids):
            cluster_items_idx = np.where(clusters == cluster_id)[0]
            num_selected_items = norm_cluster_num_item_list[i]
            random.shuffle(cluster_items_idx)
            selected_items.extend(item_list[idx] for idx in cluster_items_idx[:num_selected_items])
        return selected_items, None


class QueryClust(PruneBase):
    """
    Select items through clustering with inits that imply each label.
    """

    def base(self, ratio, num_centers, labels, database_keys, item_list, source):
        center_dict = {i: None for i in range(1, num_centers)}
        for item in item_list:
            for anno in item.annotations:
                if isinstance(anno, Label):
                    label_ = anno.label
                    if center_dict.get(label_) is None:
                        center_dict[label_] = item
            if all(center_dict.values()):
                break

        item_id_list = [item.id.split("/")[-1] for item in item_list]
        centroids = [
            database_keys[item_id_list.index(item.id)] for item in center_dict.values() if item
        ]
        kmeans = KMeans(
            n_clusters=num_centers, n_init=1, init=np.stack(centroids, axis=0), random_state=0
        )

        clusters = kmeans.fit_predict(database_keys)
        cluster_centers = kmeans.cluster_centers_
        cluster_ids, cluster_num_item_list = np.unique(clusters, return_counts=True)

        norm_cluster_num_item_list = match_num_item_for_cluster(
            ratio, len(database_keys), cluster_num_item_list
        )

        selected_items = []
        dist_tuples = []
        for i, cluster_id in enumerate(cluster_ids):
            cluster_center = cluster_centers[cluster_id]
            cluster_items_idx = np.where(clusters == cluster_id)[0]
            num_selected_item = norm_cluster_num_item_list[i]

            cluster_items = database_keys[cluster_items_idx]
            dist = calculate_hamming(cluster_center, cluster_items)
            ind = np.argsort(dist)
            item_idx_list = cluster_items_idx[ind]
            for i, idx in enumerate(item_idx_list[:num_selected_item]):
                selected_items.append(item_list[idx])
                dist_tuples.append(
                    (cluster_id, item_list[idx].id, item_list[idx].subset, dist[ind][i])
                )
        return selected_items, dist_tuples


class Entropy(PruneBase):
    """
    Select items through clustering and choose them based on label entropy in each cluster.
    """

    def base(self, ratio, num_centers, labels, database_keys, item_list, source):
        kmeans = KMeans(n_clusters=num_centers, random_state=0)
        clusters = kmeans.fit_predict(database_keys)

        cluster_ids, cluster_num_item_list = np.unique(clusters, return_counts=True)
        norm_cluster_num_item_list = match_num_item_for_cluster(
            ratio, len(database_keys), cluster_num_item_list
        )

        selected_item_indexes = []
        for cluster_id, num_selected_item in zip(cluster_ids, norm_cluster_num_item_list):
            cluster_items_idx = np.where(clusters == cluster_id)[0]

            cluster_classes = np.array(labels)[cluster_items_idx]
            _, inv, cnts = np.unique(cluster_classes, return_inverse=True, return_counts=True)
            weights = 1 / cnts
            probs = weights[inv]
            probs /= probs.sum()

            choices = np.random.choice(len(inv), size=num_selected_item, p=probs, replace=False)
            selected_item_indexes.extend(cluster_items_idx[choices])

        selected_items = np.array(item_list)[selected_item_indexes].tolist()
        return selected_items, None


class NDRSelect(PruneBase):
    """
    Select items based on NDR among each subset.
    """

    def base(self, ratio, num_centers, labels, database_keys, item_list, source):
        subset_lists = list(source.subsets().keys())

        selected_items = []
        for subset_ in subset_lists:
            subset_len = len(source.get_subset(subset_))
            num_selected_subset_item = math.ceil(subset_len * (1 - ratio))
            ndr_result = ndr.NDR(source, working_subset=subset_, num_cut=num_selected_subset_item)
            selected_items.extend(ndr_result.get_subset(subset_))

        return selected_items, None


class Prune(HashInference):
    def __init__(
        self,
        dataset: Dataset,
        cluster_method: str = "random",
        hash_type: str = "img",
    ) -> None:
        """
        Prune make a representative and manageable subset.
        """
        self._dataset = dataset
        self._cluster_method = cluster_method
        self._hash_type = hash_type

        self._model = None
        self._text_model = None
        self._num_centers = None

        self._database_keys = None
        self._item_list = []
        self._labels = []

        self._prepare_data()

    def _prepare_data(self):
        if self._hash_type == "txt":
            category_dict = self._prompting()

        if self._cluster_method == "random":
            self._item_list = list(self._dataset)
            return

        datasets_to_infer = select_uninferenced_dataset(self._dataset)
        datasets = self._compute_hash_key([self._dataset], [datasets_to_infer])[0]

        for category in datasets.categories().values():
            if isinstance(category, LabelCategories):
                self._num_centers = len(category._indices.keys())

        for item in datasets:
            for annotation in item.annotations:
                if isinstance(annotation, Label):
                    self._labels.append(annotation.label)
                if isinstance(annotation, HashKey):
                    hash_key = annotation.hash_key
                    if self._hash_type == "txt":
                        inputs = category_dict.get(str(item.annotations[0].label))
                        if isinstance(inputs, List):
                            inputs = " ".join(inputs)
                        hash_key_txt = self.text_model.infer_text(inputs).hash_key
                        hash_key = np.concatenate([hash_key, hash_key_txt])
                    hash_key = np.unpackbits(hash_key, axis=-1)
                    if self._database_keys is None:
                        self._database_keys = hash_key.reshape(1, -1)
                    else:
                        self._database_keys = np.concatenate(
                            (self._database_keys, hash_key.reshape(1, -1)), axis=0
                        )
            self._item_list.append(item)

    def _prompting(self):
        category_dict = {}
        detected_format = self._dataset.format
        template = format_templates.get(detected_format, templates)
        for label in list(self._dataset.categories().values())[0]._indices.keys():
            category_dict[label] = [temp.format(label) for temp in template]
        return category_dict

    def get_pruned(self, ratio: float = 0.5) -> Dataset:
        method = {
            "random": RandomSelect,
            "cluster_random": ClusteredRandom,
            "centroid": Centroid,
            "query_clust": QueryClust,
            "entropy": Entropy,
            "ndr": NDRSelect,
        }

        prune_method = method[self._cluster_method]()
        selected_items, dist_tuples = prune_method.base(
            ratio=ratio,
            num_centers=self._num_centers,
            labels=self._labels,
            database_keys=self._database_keys,
            item_list=self._item_list,
            source=self._dataset,
        )

        result_dataset = Dataset(media_type=self._dataset.media_type())
        result_dataset._source_path = self._dataset._source_path
        result_dataset.define_categories(self._dataset.categories())
        for item in selected_items:
            result_dataset.put(item)

        if dist_tuples:
            for center, id_, subset_, d in dist_tuples:
                log.info(f"item {id_} of subset {subset_} has distance {d} for cluster {center}")

        log.info(f"Pruned dataset with {ratio} from {len(self._dataset)} to {len(result_dataset)}")
        return result_dataset



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import glob
import logging as log
import os
import os.path as osp
import re
from typing import Optional

import numpy as np

from datumaro.components.annotation import AnnotationType, CompiledMask, LabelCategories, Mask
from datumaro.components.dataset_base import DatasetBase, DatasetItem
from datumaro.components.errors import InvalidAnnotationError
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.image import IMAGE_EXTENSIONS, find_images, lazy_image, load_image
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file


class Ade20k2017Path:
    MASK_PATTERN = re.compile(
        r""".+_seg
        | .+_parts_\d+
    """,
        re.VERBOSE,
    )


class Ade20k2017Base(DatasetBase):
    def __init__(self, path: str, *, ctx: Optional[ImportContext] = None):
        if not osp.isdir(path):
            raise NotADirectoryError(errno.ENOTDIR, "Can't find dataset directory", path)

        # exclude dataset meta file
        subsets = [subset for subset in os.listdir(path) if osp.splitext(subset)[-1] != ".json"]
        if len(subsets) < 1:
            raise FileNotFoundError(errno.ENOENT, "Can't find subsets in directory", path)

        super().__init__(subsets=sorted(subsets), ctx=ctx)
        self._path = path

        self._items = []
        self._categories = {}

        if has_meta_file(self._path):
            self._categories = {
                AnnotationType.label: LabelCategories.from_iterable(
                    parse_meta_file(self._path).keys()
                )
            }

        for subset in self._subsets:
            self._load_items(subset)

    def __iter__(self):
        return iter(self._items)

    def categories(self):
        return self._categories

    def _load_items(self, subset):
        labels = self._categories.setdefault(AnnotationType.label, LabelCategories())
        path = osp.join(self._path, subset)

        images = [i for i in find_images(path, recursive=True)]

        for image_path in sorted(images):
            item_id = osp.splitext(osp.relpath(image_path, path))[0]

            if Ade20k2017Path.MASK_PATTERN.fullmatch(osp.basename(item_id)):
                continue

            item_annotations = []

            item_info = self._load_item_info(image_path)
            for item in item_info:
                label_idx = labels.find(item["label_name"])[0]
                if label_idx is None:
                    labels.add(item["label_name"])

            mask_path = osp.splitext(image_path)[0] + "_seg.png"
            if not osp.isfile(mask_path):
                log.warning("Can't find mask for image: %s" % image_path)

            part_level = 0
            max_part_level = max([p["part_level"] for p in item_info])
            for part_level in range(max_part_level + 1):
                if not osp.exists(mask_path):
                    log.warning("Can`t find part level %s mask for %s" % (part_level, image_path))
                    continue

                mask = lazy_image(mask_path, loader=self._load_instance_mask)
                mask = CompiledMask(instance_mask=mask)

                for v in item_info:
                    if v["part_level"] != part_level:
                        continue

                    label_id = labels.find(v["label_name"])[0]
                    instance_id = v["id"]
                    attributes = {k: True for k in v["attributes"]}

                    item_annotations.append(
                        Mask(
                            label=label_id,
                            image=mask.lazy_extract(instance_id),
                            id=instance_id,
                            attributes=attributes,
                            z_order=part_level,
                            group=instance_id,
                        )
                    )

                mask_path = osp.splitext(image_path)[0] + "_parts_%s.png" % (part_level + 1)

            self._items.append(
                DatasetItem(
                    item_id,
                    subset=subset,
                    media=Image.from_file(path=image_path),
                    annotations=item_annotations,
                )
            )

    def _load_item_info(self, path):
        attr_path = osp.splitext(path)[0] + "_atr.txt"
        if not osp.isfile(attr_path):
            raise FileNotFoundError(
                errno.ENOENT, "Can't find annotation file for image %s" % path, attr_path
            )

        item_info = []
        with open(attr_path, "r", encoding="utf-8") as f:
            for line in f:
                columns = [s.strip() for s in line.split("#")]
                if len(columns) != 6:
                    raise InvalidAnnotationError("Invalid line in %s" % attr_path)
                if columns[5][0] != '"' or columns[5][-1] != '"':
                    raise InvalidAnnotationError(
                        "Attributes column are expected \
                        in double quotes, file %s"
                        % attr_path
                    )
                attributes = [s.strip() for s in columns[5][1:-1].split(",") if s]

                item_info.append(
                    {
                        "id": int(columns[0]),
                        "part_level": int(columns[1]),
                        "occluded": int(columns[2]),
                        "label_name": columns[4],
                        "attributes": attributes,
                    }
                )

        return item_info

    @staticmethod
    def _load_instance_mask(path):
        mask = load_image(path)
        _, instance_mask = np.unique(mask[:, :, 0], return_inverse=True)
        instance_mask = instance_mask.reshape(mask[:, :, 0].shape)
        return instance_mask


class Ade20k2017Importer(Importer):
    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        context.require_file("*/**/*_atr.txt")

    @classmethod
    def find_sources(cls, path):
        for i in range(5):
            for i in glob.iglob(osp.join(path, *("*" * i))):
                if osp.splitext(i)[1].lower() in IMAGE_EXTENSIONS:
                    return [
                        {
                            "url": path,
                            "format": Ade20k2017Base.NAME,
                        }
                    ]
        return []



# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import os
import os.path as osp
from typing import Dict, List, Optional, Tuple, Type, Union

from datumaro.components.annotation import AnnotationType, Categories, Tabular, TabularCategories
from datumaro.components.dataset_base import DatasetBase, DatasetItem
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Table, TableDtype, TableRow
from datumaro.util.os_util import find_files

# Only supports '.csv' extention.
TABULAR_EXTENSIONS = [
    "csv",
]


class TabularDataBase(DatasetBase):
    NAME = "tabular"

    def __init__(
        self,
        path: str,
        *,
        target: Optional[Union[str, List[str]]] = None,
        dtype: Optional[Dict[str, Type[TableDtype]]] = None,
        ctx: Optional[ImportContext] = None,
    ) -> None:
        """
        Read and compose a tabular dataset.
        The file name of each '.csv' file is regarded as subset.

        Args:
            path (str) : Path to a tabular dataset. (csv file or folder contains csv files).
            target (optional, str or list(str)) : Target column or list of target columns.
                If this is not specified (None), the last column is regarded as a target column.
                In case of a dataset with no targets, give an empty list as a parameter.
            dtype (optional, dict(str,str)) : Dictionay of column name -> type str ('str', 'int', or 'float').
                This can be used when automatic type inferencing is failed.
        """
        paths: List[str] = []
        if osp.isfile(path):
            paths.append(path)
        else:
            for path in find_files(path, TABULAR_EXTENSIONS):
                paths.append(path)

        if not paths:
            raise FileNotFoundError(errno.ENOENT, "Can't find tabular files", path)

        super().__init__(media_type=TableRow, ctx=ctx)

        self._infos = {"path": path}
        self._items, self._categories = self._parse(paths, target, dtype)

    def _parse(
        self,
        paths: List[str],
        target: Optional[Union[str, List[str]]] = None,
        dtype: Optional[Dict[str, Type[TableDtype]]] = None,
    ) -> Tuple[List[DatasetItem], Dict[AnnotationType, Categories]]:
        """
        parse tabular files. Each file is regarded as a subset.

        Args:
            paths (list(str)) : A list of paths to tabular data files(csv files).
            target (optional, str or list) : Target column or list of target columns.
                If this is not specified (None), the last column is regarded as a target column.
                In case of a dataset with no targets, give an empty list as a parameter.
            dtype (optional, dict(str,str)) : Dictionay of column name -> type str ('str', 'int', or 'float').
                This can be used when automatic type inferencing is failed.

        Returns:
            list (DatasetItem): dataset items
            dict (AnnotationType, Categories): categories info
        """
        # assert paths
        items: List[DatasetItem] = []
        categories: TabularCategories = TabularCategories()

        for path in paths:
            table = Table.from_csv(path, dtype=dtype)

            targets: List[str] = []
            if target is None:
                targets.append(table.columns[-1])  # last column
            elif isinstance(target, str):
                if target in table.columns:  # add valid column name only
                    targets.append(target)
            elif isinstance(target, list):  # add valid column names only
                for t in target:
                    if t in table.columns:
                        targets.append(t)

            # set categories
            for target in targets:
                _, category = categories.find(target)
                target_dtype = table.dtype(target)
                if target_dtype == str:
                    labels = set(table.features(target, unique=True))
                    if category is None:
                        categories.add(target, target_dtype, labels)
                    else:  # update labels if they are different.
                        category.labels.union(labels)
                elif target_dtype in [int, float]:
                    # 'int' can be categorical, but we don't know this unless user gives information.
                    if category is None:
                        categories.add(target, target_dtype)
                else:
                    raise TypeError(
                        f"Unsupported type '{target_dtype}' for target column '{target}'."
                    )

            # load annotations
            subset = osp.splitext(osp.basename(path))[0]
            row: TableRow
            for row in table:  # type: TableRow
                id = f"{row.index}@{subset}"
                ann = [Tabular(values=row.data(targets))] if targets else None
                item = DatasetItem(
                    id=id,
                    subset=subset,
                    media=row,
                    annotations=ann,
                )
                items.append(item)

        return items, {AnnotationType.tabular: categories}

    def categories(self):
        return self._categories

    def __iter__(self):
        yield from self._items


class TabularDataImporter(Importer):
    """
    Import a tabular dataset.
    Each '.csv' file is regarded as a subset.
    """

    NAME = "tabular"

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--target",
            type=lambda x: x.split(","),
            help="Target column or list of target columns. (ex. 'class', 'class,breed') (default:None) "
            "If this is not specified (None), the last column is regarded as a target column."
            "In case of a dataset with no targets, give an empty list as a parameter.",
        )
        parser.add_argument(
            "--dtype",
            type=lambda x: {k: v for k, v in (map.split(":") for map in x.split(","))},
            help="Type information for a column. (ex. 'date:str,x:int') (default:None) "
            "This can be used when automatic type inferencing is failed",
        )
        return parser

    @classmethod
    def find_sources(cls, path):
        if not osp.isdir(path):
            ext = osp.splitext(path)[1][1:]  # exclude "."
            if ext in TABULAR_EXTENSIONS:
                return [{"url": path, "format": TabularDataBase.NAME}]
        else:
            for _ in find_files(path, TABULAR_EXTENSIONS):  # find 1 depth only.
                return [{"url": path, "format": TabularDataBase.NAME}]
        return []


class TabularDataExporter(Exporter):
    """
    Export a tabular dataset.
    This will save each subset into a '.csv' file regardless of 'save_media' value
    """

    NAME = "tabular"
    EXPORT_EXT = ".csv"
    DEFAULT_IMAGE_EXT = ".jpg"  # just to avoid assert error.

    def _apply_impl(self):
        extractor = self._extractor

        if extractor.media_type() and not issubclass(extractor.media_type(), TableRow):
            raise MediaTypeError("Media type is not a table.")

        # we don't check self._save_media.
        # regardless of the value, we always save media(csv) file.

        os.makedirs(self._save_dir, exist_ok=True)

        for sname in extractor.subsets():
            subset = extractor.get_subset(sname)
            path = osp.join(self._save_dir, sname + self.EXPORT_EXT)
            list_of_dicts: List[Dict[str, TableDtype]] = list()
            for item in subset:
                dicts = item.media.data()
                for ann in item.annotations:
                    if isinstance(ann, Tabular):
                        dicts.update(ann.values)  # update value
                list_of_dicts.append(dicts)

            table = Table.from_list(list_of_dicts)
            table.save(path)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
from typing import Dict, List, Optional

from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer
from datumaro.errors import DatasetImportError
from datumaro.util import parse_json


class SegmentAnythingImporter(Importer):
    _N_JSON_TO_TEST = 10

    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> Optional[FormatDetectionConfidence]:
        # test maximum 10 annotation files only
        ctr = 0
        for file in context.require_files_iter("*.json"):
            ctr += 1
            with context.probe_text_file(
                file, "Annotation format is not Segmentat-Anything format", is_binary_file=True
            ) as f:
                anno = parse_json(f.read())
                if (
                    set(anno.keys()) != {"annotations", "image"}
                    or (
                        set(anno["image"].keys())
                        != {
                            "image_id",
                            "width",
                            "height",
                            "file_name",
                        }
                    )
                    or (
                        anno["annotations"]
                        and not {"id", "segmentation", "bbox"}.issubset(set(anno["annotations"][0]))
                    )
                ):
                    raise DatasetImportError
            if ctr > cls._N_JSON_TO_TEST:
                break

    @classmethod
    def find_sources(cls, path) -> List[Dict]:
        if not os.path.isdir(path):
            return []
        return [{"url": path, "format": cls.NAME}]



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import importlib
import os
import os.path as osp
import shlex
import weakref
from contextlib import contextmanager
from enum import Enum, auto
from io import BytesIO, IOBase
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Tuple, Union

import numpy as np

from datumaro.components.crypter import NULL_CRYPTER, Crypter
from datumaro.components.errors import DatumaroError

try:
    # Introduced in 1.20
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any


class _IMAGE_BACKENDS(Enum):
    cv2 = auto()
    PIL = auto()


_IMAGE_BACKEND = None
_image_loading_errors = (FileNotFoundError,)
try:
    importlib.import_module("cv2")
    _IMAGE_BACKEND = _IMAGE_BACKENDS.cv2
except ModuleNotFoundError:
    import PIL

    _IMAGE_BACKEND = _IMAGE_BACKENDS.PIL
    _image_loading_errors = (*_image_loading_errors, PIL.UnidentifiedImageError)

from datumaro.util.image_cache import ImageCache
from datumaro.util.os_util import find_files


def load_image(path: str, dtype: DTypeLike = np.uint8, crypter: Crypter = NULL_CRYPTER):
    """
    Reads an image in the HWC Grayscale/BGR(A) [0; 255] format (default dtype is uint8).
    """

    if _IMAGE_BACKEND == _IMAGE_BACKENDS.cv2:
        # cv2.imread does not support paths that are not representable
        # in the locale encoding on Windows, so we read the image bytes
        # ourselves.

        with open(path, "rb") as f:
            image_bytes = crypter.decrypt(f.read())

        return decode_image(image_bytes, dtype=dtype)
    elif _IMAGE_BACKEND == _IMAGE_BACKENDS.PIL:
        from PIL import Image

        if not crypter.is_null_crypter:
            raise DatumaroError("PIL backend should have crypter=NullCrypter.")

        image = Image.open(path)
        image = np.asarray(image, dtype=dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image = np.array(image)  # Release read-only
            image[:, :, :3] = image[:, :, 2::-1]  # RGB to BGR
    else:
        raise NotImplementedError()

    assert len(image.shape) in {2, 3}
    if len(image.shape) == 3:
        assert image.shape[2] in {3, 4}
    return image


def copyto_image(
    src: Union[str, IOBase], dst: Union[str, IOBase], src_crypter: Crypter, dst_crypter: Crypter
) -> None:
    if src_crypter == dst_crypter and src == dst:
        return

    @contextmanager
    def _open(fp, mode):
        was_file = False
        if not isinstance(fp, IOBase):
            was_file = True
            fp = open(fp, mode)
        yield fp
        if was_file:
            fp.close()

    with _open(src, "rb") as src_fp:
        _bytes = src_crypter.decrypt(src_fp.read())

    with _open(dst, "wb") as dst_fp:
        dst_fp.write(dst_crypter.encrypt(_bytes))


def save_image(
    dst: Union[str, IOBase],
    image: np.ndarray,
    ext: Optional[str] = None,
    create_dir: bool = False,
    dtype: DTypeLike = np.uint8,
    crypter: Crypter = NULL_CRYPTER,
    **kwargs,
) -> None:
    # NOTE: Check destination path for existence
    # OpenCV silently fails if target directory does not exist
    if isinstance(dst, IOBase):
        ext = ext if ext else ".png"
    else:
        dst_dir = osp.dirname(dst)
        if dst_dir:
            if create_dir:
                os.makedirs(dst_dir, exist_ok=True)
            elif not osp.isdir(dst_dir):
                raise FileNotFoundError("Directory does not exist: '%s'" % dst_dir)
        # file extension and actual encoding can be differ
        ext = ext if ext else osp.splitext(dst)[1]

    if not kwargs:
        kwargs = {}

    # NOTE: OpenCV documentation says "If the image format is not supported,
    # the image will be converted to 8-bit unsigned and saved that way".
    # Conversion from np.int32 to np.uint8 is not working properly
    backend = _IMAGE_BACKEND
    if dtype == np.int32:
        backend = _IMAGE_BACKENDS.PIL
    if backend == _IMAGE_BACKENDS.cv2:
        # cv2.imwrite does not support paths that are not representable
        # in the locale encoding on Windows, so we write the image bytes
        # ourselves.

        image_bytes = encode_image(image, ext, dtype=dtype, **kwargs)

        if isinstance(dst, str):
            with open(dst, "wb") as f:
                f.write(crypter.encrypt(image_bytes))
        else:
            dst.write(crypter.encrypt(image_bytes))
    elif backend == _IMAGE_BACKENDS.PIL:
        from PIL import Image

        if ext.startswith("."):
            ext = ext[1:]

        if not crypter.is_null_crypter:
            raise DatumaroError("PIL backend should have crypter=NullCrypter.")

        params = {}
        params["quality"] = kwargs.get("jpeg_quality")
        if kwargs.get("jpeg_quality") == 100:
            params["subsampling"] = 0

        image = image.astype(dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image[:, :, :3] = image[:, :, 2::-1]  # BGR to RGB
        image = Image.fromarray(image)
        image.save(dst, format=ext, **params)
    else:
        raise NotImplementedError()


def encode_image(image: np.ndarray, ext: str, dtype: DTypeLike = np.uint8, **kwargs) -> bytes:
    if not kwargs:
        kwargs = {}

    if _IMAGE_BACKEND == _IMAGE_BACKENDS.cv2:
        import cv2

        params = []

        if not ext.startswith("."):
            ext = "." + ext

        if ext.upper() in (".JPG", ".JPEG"):
            params = [int(cv2.IMWRITE_JPEG_QUALITY), kwargs.get("jpeg_quality", 75)]

        image = image.astype(dtype)
        success, result = cv2.imencode(ext, image, params=params)
        if not success:
            raise Exception("Failed to encode image to '%s' format" % (ext))
        return result.tobytes()
    elif _IMAGE_BACKEND == _IMAGE_BACKENDS.PIL:
        from PIL import Image

        if ext.startswith("."):
            ext = ext[1:]

        params = {}
        params["quality"] = kwargs.get("jpeg_quality")
        if kwargs.get("jpeg_quality") == 100:
            params["subsampling"] = 0

        image = image.astype(dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image[:, :, :3] = image[:, :, 2::-1]  # BGR to RGB
        image = Image.fromarray(image)
        with BytesIO() as buffer:
            image.save(buffer, format=ext, **params)
            return buffer.getvalue()
    else:
        raise NotImplementedError()


def decode_image(image_bytes: bytes, dtype: DTypeLike = np.uint8) -> np.ndarray:
    if _IMAGE_BACKEND == _IMAGE_BACKENDS.cv2:
        import cv2

        image = np.frombuffer(image_bytes, dtype=np.uint8)
        image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
        image = image.astype(dtype)
    elif _IMAGE_BACKEND == _IMAGE_BACKENDS.PIL:
        from PIL import Image

        image = Image.open(BytesIO(image_bytes))
        image = np.asarray(image, dtype=dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image = np.array(image)  # Release read-only
            image[:, :, :3] = image[:, :, 2::-1]  # RGB to BGR
    else:
        raise NotImplementedError()

    assert len(image.shape) in {2, 3}
    if len(image.shape) == 3:
        assert image.shape[2] in {3, 4}
    return image


IMAGE_EXTENSIONS = {
    ".jpg",
    ".jpeg",
    ".jpe",
    ".jp2",
    ".png",
    ".bmp",
    ".dib",
    ".tif",
    ".tiff",
    ".tga",
    ".webp",
    ".pfm",
    ".sr",
    ".ras",
    ".exr",
    ".hdr",
    ".pic",
    ".pbm",
    ".pgm",
    ".ppm",
    ".pxm",
    ".pnm",
}


def find_images(
    dirpath: str,
    exts: Union[str, Iterable[str]] = None,
    recursive: bool = False,
    max_depth: Optional[int] = None,
    min_depth: Optional[int] = None,
) -> Iterator[str]:
    yield from find_files(
        dirpath,
        exts=exts or IMAGE_EXTENSIONS,
        recursive=recursive,
        max_depth=max_depth,
        min_depth=min_depth,
    )


def is_image(path: str) -> bool:
    trunk, ext = osp.splitext(osp.basename(path))
    return trunk and ext.lower() in IMAGE_EXTENSIONS and osp.isfile(path)


class lazy_image:
    def __init__(
        self,
        path: str,
        loader: Callable[[str], np.ndarray] = None,
        cache: Union[bool, ImageCache] = True,
        crypter: Crypter = NULL_CRYPTER,
    ) -> None:
        """
        Cache:
            - False: do not cache
            - True: use the global cache
            - ImageCache instance: an object to be used as cache
        """

        self._custom_loader = True

        if loader is None:
            loader = load_image
            self._custom_loader = False

        self._path = path
        self._loader = loader

        assert isinstance(cache, (ImageCache, bool))
        self._cache = cache
        self._crypter = crypter

    def __call__(self) -> np.ndarray:
        image = None
        cache_key = weakref.ref(self)

        cache = self._get_cache()
        if cache is not None:
            image = cache.get(cache_key)

        if image is None:
            image = (
                self._loader(self._path)
                if self._custom_loader
                else self._loader(self._path, crypter=self._crypter)
            )
            if cache is not None:
                cache.push(cache_key, image)
        return image

    def _get_cache(self) -> Optional[ImageCache]:
        if self._cache is True:
            cache = ImageCache.get_instance()
        elif self._cache is False:
            cache = None
        else:
            cache = self._cache
        return cache


ImageMeta = Dict[str, Tuple[int, int]]
"""filename -> height, width"""

DEFAULT_IMAGE_META_FILE_NAME = "images.meta"


def load_image_meta_file(image_meta_path: str) -> ImageMeta:
    """
    Loads image metadata from a file with the following format:

        <image name 1> <height 1> <width 1>

        <image name 2> <height 2> <width 2>

        ...

    Shell-like comments and quoted fields are allowed.

    This can be useful to support datasets in which image dimensions are
    required to interpret annotations.
    """
    assert isinstance(image_meta_path, str)

    if not osp.isfile(image_meta_path):
        raise FileNotFoundError("Can't read image meta file '%s'" % image_meta_path)

    image_meta = {}

    with open(image_meta_path, encoding="utf-8") as f:
        for line in f:
            fields = shlex.split(line, comments=True)
            if not fields:
                continue

            # ignore extra fields, so that the format can be extended later
            image_name, h, w = fields[:3]
            image_meta[image_name] = (int(h), int(w))

    return image_meta


def save_image_meta_file(image_meta: ImageMeta, image_meta_path: str) -> None:
    """
    Saves image_meta to the path specified by image_meta_path in the format
    defined in load_image_meta_file's documentation.
    """

    assert isinstance(image_meta_path, str)

    with open(image_meta_path, "w", encoding="utf-8") as f:
        # Add a comment about file syntax
        print("# <image name> <height> <width>", file=f)
        print("", file=f)

        for image_name, (height, width) in image_meta.items():
            print(shlex.quote(image_name), height, width, file=f)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import datetime
import os
import platform
import re
import struct
import tempfile
from copy import deepcopy
from functools import partial
from multiprocessing.pool import ApplyResult, Pool
from shutil import move, rmtree
from typing import Any, Callable, Dict, Optional, Union

import pyarrow as pa
import pytz

from datumaro.components.crypter import NULL_CRYPTER
from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME, DatasetItem, IDataset
from datumaro.components.errors import DatumaroError
from datumaro.components.exporter import ExportContext, ExportContextComponent, Exporter
from datumaro.plugins.data_formats.datumaro.exporter import _SubsetWriter as __SubsetWriter
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import DictMapper
from datumaro.util.file_utils import to_bytes

from .format import DatumaroArrow
from .mapper.dataset_item import DatasetItemMapper
from .mapper.media import ImageMapper


class PathNormalizer:
    NORMALIZER = {r":[^/\\]": r"꞉"}
    UNNORMALIZER = {r"꞉": r":"}

    @classmethod
    def normalize(cls, path: str):
        for s, t in cls.NORMALIZER.items():
            path = re.sub(s, t, path)
        return path

    @classmethod
    def unnormalize(cls, path: str):
        for s, t in cls.UNNORMALIZER.items():
            path = re.sub(s, t, path)
        return path


class _SubsetWriter(__SubsetWriter):
    def __init__(
        self,
        context: Exporter,
        export_context: ExportContextComponent,
        subset: str,
        ctx: ExportContext,
        max_chunk_size: int = 1000,
        num_shards: int = 1,
        max_shard_size: Optional[int] = None,
    ):
        super().__init__(
            context=context,
            subset=subset,
            ann_file="",
            export_context=export_context,
        )
        self._schema = deepcopy(DatumaroArrow.SCHEMA)
        self._writers = []
        self._fnames = []
        self._max_chunk_size = max_chunk_size
        self._num_shards = num_shards
        self._max_shard_size = max_shard_size
        self._ctx = ctx

        self._data = {
            "items": [],
            "infos": {},
            "categories": {},
            "media_type": None,
            "built_time": str(datetime.datetime.now(pytz.utc)),
            "source_path": self.export_context.source_path,
            "encoding_scheme": self.export_context._image_ext,
            "version": str(DatumaroArrow.VERSION),
            "signature": DatumaroArrow.SIGNATURE,
        }

    def add_infos(self, infos):
        if self._writers:
            raise ValueError("Writer has been initialized.")
        super().add_infos(infos)
        self._data["infos"] = DictMapper.forward(self.infos)

    def add_categories(self, categories):
        if self._writers:
            raise ValueError("Writer has been initialized.")
        super().add_categories(categories)
        self._data["categories"] = DictMapper.forward(self.categories)

    def add_media_type(self, media_type):
        if self._writers:
            raise ValueError("Writer has been initialized.")
        self._data["media_type"] = struct.pack("<I", int(media_type))

    def init_schema(self):
        self._schema = self._schema.with_metadata(
            {k: v for k, v in self._data.items() if k != "items"}
        )

    def _init_writer(self, idx: int):
        # TODO:
        # `arrow_writer.close()` does not take any effect. It seems a bug in pyarrow
        # In linux-like system it is fine to rename a file opend but
        # it is not in windows.
        # As a workaround, we do not use a template in windows
        if platform.system() != "Windows":
            f_name = os.path.join(
                self.export_context.save_dir,
                self._subset + "-{idx" + str(idx) + ":0{width}d}-of-{total:0{width}d}.arrow",
            )
            f_name = PathNormalizer.normalize(f_name)
        else:
            f_name = os.path.join(
                self.export_context.save_dir, self._subset + f"-{idx:03d}-of-{0:03d}.arrow"
            )
        return pa.RecordBatchStreamWriter(f_name, self._schema), f_name

    def add_item(self, item: DatasetItem, pool: Optional[Pool] = None):
        if pool is not None:
            self.items.append(
                pool.apply_async(
                    self.add_item_impl,
                    (
                        item,
                        self.export_context,
                    ),
                )
            )
        else:
            self.items.append(partial(self.add_item_impl, item, self.export_context))

    @staticmethod
    def add_item_impl(
        item: DatasetItem,
        context: ExportContextComponent,
    ) -> Dict[str, Any]:
        item = DatasetItemMapper.forward(item, media={"encoder": context._image_ext})

        if item["media"].get("bytes", None) is not None:
            # truncate source path since the media is embeded in arrow
            path = item["media"].get("path")
            if path is not None:
                item["media"]["path"] = path.replace(context.source_path, "")
        return item

    def _write(self, batch, max_chunk_size):
        if max_chunk_size == 0:
            max_chunk_size = self._max_chunk_size
        pa_table = pa.Table.from_arrays(batch, schema=self._schema)
        idx = getattr(self, "__writer_idx", 0)

        if self._max_shard_size is not None:
            nbytes = pa_table.nbytes
            cur_nbytes = getattr(self, "__writer_nbytes", 0)
            free_nbytes = self._max_shard_size - cur_nbytes
            if free_nbytes < nbytes:
                if cur_nbytes == 0:
                    raise DatumaroError(
                        "'max_chunk_size' exceeded 'max_shard_size'. "
                        "Please consider increasing 'max_shard_size' or deceasing 'max_chunk_size'."
                    )
                self._writers[idx].close()
                idx += 1
                setattr(self, "__writer_idx", idx)
                setattr(self, "__writer_nbytes", nbytes)
            else:
                setattr(self, "__writer_nbytes", cur_nbytes + nbytes)
        else:
            setattr(self, "__writer_idx", (idx + 1) % self._num_shards)

        if len(self._writers) <= idx:
            writer, fname = self._init_writer(idx)
            self._writers.append(writer)
            self._fnames.append(fname)
        self._writers[idx].write_table(pa_table, max_chunk_size)

    def write(self, max_chunk_size: Optional[int] = None, pool: Optional[Pool] = None):
        if max_chunk_size is None:
            max_chunk_size = self._max_chunk_size

        if len(self.items) < max_chunk_size:
            return

        batch = [[] for _ in self._schema.names]
        for item in self._ctx.progress_reporter.iter(
            self.items, desc=f"Building arrow for {self._subset}"
        ):
            if isinstance(item, ApplyResult):
                item = item.get(timeout=DatumaroArrow.MP_TIMEOUT)
            if isinstance(item, partial):
                item = item()
            for j, name in enumerate(self._schema.names):
                batch[j].append(item[name])

            if len(batch[0]) >= max_chunk_size:
                self._write(batch, max_chunk_size)
                batch = [[] for _ in self._schema.names]
        if len(batch[0]) > 0:
            self._write(batch, max_chunk_size)

        self._data["items"] = []

    def done(self):
        total = len(self._fnames)
        width = len(str(total))
        for idx, (fname, writer) in enumerate(zip(self._fnames, self._writers)):
            writer.close()
            if platform.system() != "Windows":
                template = PathNormalizer.unnormalize(fname)
                placeholders = {"width": width, "total": total, f"idx{idx}": idx}
                new_fname = template.format(**placeholders)
                if os.path.exists(new_fname):
                    os.remove(new_fname)
                os.rename(fname, new_fname)


class ArrowExporter(Exporter):
    AVAILABLE_IMAGE_EXTS = ImageMapper.AVAILABLE_SCHEMES
    DEFAULT_IMAGE_EXT = ImageMapper.AVAILABLE_SCHEMES[0]

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)

        # '--image-ext' would be used in a different way for arrow foramt
        _actions = []
        for action in parser._actions:
            if action.dest != "image_ext":
                _actions.append(action)
        parser._actions = _actions
        parser._option_string_actions.pop("--image-ext")

        parser.add_argument(
            "--image-ext",
            default=None,
            help=f"Image encoding scheme. (default: {cls.DEFAULT_IMAGE_EXT})",
            choices=cls.AVAILABLE_IMAGE_EXTS,
        )

        parser.add_argument(
            "--max-chunk-size",
            type=int,
            default=1000,
            help="The maximum chunk size. (default: %(default)s)",
        )

        parser.add_argument(
            "--num-shards",
            type=int,
            default=1,
            help="The number of shards to export. "
            "'--num-shards' and '--max-shard-size' are  mutually exclusive. "
            "(default: %(default)s)",
        )

        parser.add_argument(
            "--max-shard-size",
            type=str,
            default=None,
            help="The maximum size of each shard. "
            "(e.g. 7KB = 7 * 2^10, 3MB = 3 * 2^20, and 2GB = 2 * 2^30). "
            "'--num-shards' and '--max-shard-size' are  mutually exclusive. "
            "(default: %(default)s)",
        )

        parser.add_argument(
            "--num-workers",
            type=int,
            default=0,
            help="The number of multi-processing workers for export. "
            "If num_workers = 0, do not use multiprocessing. (default: %(default)s)",
        )

        return parser

    def create_writer(self, subset: str, ctx: ExportContext) -> _SubsetWriter:
        export_context = ExportContextComponent(
            save_dir=self._save_dir,
            save_media=self._save_media,
            images_dir="",
            pcd_dir="",
            video_dir="",
            crypter=NULL_CRYPTER,
            image_ext=self._image_ext,
            default_image_ext=self._default_image_ext,
            source_path=os.path.abspath(self._extractor._source_path)
            if getattr(self._extractor, "_source_path")
            else None,
        )

        return _SubsetWriter(
            context=self,
            subset=subset,
            export_context=export_context,
            num_shards=self._num_shards,
            max_shard_size=self._max_shard_size,
            max_chunk_size=self._max_chunk_size,
            ctx=ctx,
        )

    def _apply_impl(self, *args, **kwargs):
        if self._num_workers == 0:
            return self._apply()

        with Pool(processes=self._num_workers) as pool:
            return self._apply(pool)

    def _apply(self, pool: Optional[Pool] = None):
        os.makedirs(self._save_dir, exist_ok=True)

        if self._split_by_subsets:
            writers = {
                subset_name: self.create_writer(subset_name, self._ctx)
                for subset_name, subset in self._extractor.subsets().items()
                if len(subset)
            }
        else:
            writers = {DEFAULT_SUBSET_NAME: self.create_writer(DEFAULT_SUBSET_NAME, self._ctx)}

        for writer in writers.values():
            writer.add_infos(self._extractor.infos())
            writer.add_categories(self._extractor.categories())
            writer.add_media_type(self._extractor.media_type()._type)
            writer.init_schema()

        for subset_name, subset in self._extractor.subsets().items():
            writer = (
                writers[subset_name] if self._split_by_subsets else writers[DEFAULT_SUBSET_NAME]
            )
            for item in subset:
                writer.add_item(item, pool=pool)

        for writer in writers.values():
            writer.write(0, pool=pool)
            writer.done()

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        # no patch supported
        with tempfile.TemporaryDirectory() as temp_dir:
            cls.convert(dataset, save_dir=temp_dir, **kwargs)
            if os.path.exists(save_dir):
                for file in os.listdir(save_dir):
                    file = os.path.join(save_dir, file)
                    if os.path.isdir(file):
                        rmtree(file)
                    else:
                        os.remove(file)
            for file in os.listdir(temp_dir):
                file_from = os.path.join(temp_dir, file)
                file_to = os.path.join(save_dir, file)
                move(file_from, file_to)

    def __init__(
        self,
        extractor: IDataset,
        save_dir: str,
        *,
        save_media: Optional[bool] = None,
        image_ext: Optional[Union[str, Callable[[str], bytes]]] = None,
        default_image_ext: Optional[str] = None,
        save_dataset_meta: bool = False,
        ctx: Optional[ExportContext] = None,
        num_workers: int = 0,
        num_shards: int = 1,
        max_shard_size: Optional[int] = None,
        max_chunk_size: int = 1000,
        **kwargs,
    ):
        super().__init__(
            extractor=extractor,
            save_dir=save_dir,
            save_media=save_media,
            image_ext=image_ext,
            default_image_ext=default_image_ext,
            save_dataset_meta=save_dataset_meta,
            ctx=ctx,
        )

        # TODO: Support a whole single file of arrow
        self._split_by_subsets = True

        if num_workers < 0:
            raise DatumaroError(
                f"num_workers should be non-negative but num_workers={num_workers}."
            )
        self._num_workers = num_workers

        if num_shards != 1 and max_shard_size is not None:
            raise DatumaroError(
                "Either one of 'num_shards' or 'max_shard_size' should be provided, not both."
            )

        if num_shards < 0:
            raise DatumaroError(f"num_shards should be non-negative but num_shards={num_shards}.")
        self._num_shards = num_shards
        self._max_shard_size = to_bytes(max_shard_size) if max_shard_size else max_shard_size

        if max_chunk_size < 0:
            raise DatumaroError(
                f"max_chunk_size should be non-negative but max_chunk_size={max_chunk_size}."
            )
        self._max_chunk_size = max_chunk_size

        if self._save_media:
            self._image_ext = (
                self._image_ext if self._image_ext is not None else self._default_image_ext
            )
        else:
            self._image_ext = "NONE"

        assert (
            self._image_ext in self.AVAILABLE_IMAGE_EXTS
        ), f"{self._image_ext} is unkonwn ext. Available exts are {self.AVAILABLE_IMAGE_EXTS}"



# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import contextlib
import logging as log
import os
import os.path as osp
import sys
from typing import Dict

from datumaro.components.environment import DEFAULT_ENVIRONMENT
from datumaro.components.extractor_tfds import (
    AVAILABLE_TFDS_DATASETS,
    TFDS_EXTRACTOR_AVAILABLE,
    TfdsDatasetRemoteMetadata,
)
from datumaro.util import dump_json
from datumaro.util.os_util import make_file_name

from ..util import MultilineFormatter
from ..util.errors import CliException
from ..util.project import generate_next_file_name


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Download a publicly available dataset",
        description="""
        Downloads a publicly available dataset and saves it in a given format.|n
        |n
        Currently, the only source of datasets is the TensorFlow Datasets
        library. Therefore, to use this command you must install TensorFlow &
        TFDS, which you can do as follows:|n
        |n
        |s|spip install datumaro[tf,tfds]|n
        |n
        To download the dataset, run "datum download run". On the other hand,
        for information about the datasets, run "datum download describe".
        """,
        formatter_class=MultilineFormatter,
    )
    subparsers = parser.add_subparsers(title="Commands")

    build_get_subparser(subparsers)
    build_describe_subparser(subparsers)


def build_get_subparser(subparsers: argparse._SubParsersAction):
    builtin_writers = sorted(DEFAULT_ENVIRONMENT.exporters)
    if TFDS_EXTRACTOR_AVAILABLE:
        available_datasets = ", ".join(f"tfds:{name}" for name in AVAILABLE_TFDS_DATASETS)
    else:
        available_datasets = "N/A (TensorFlow and/or TensorFlow Datasets " "are not installed)"

    parser = subparsers.add_parser(
        name="get",
        help="Download a publicly available dataset",
        description="""
        Supported datasets: {}|n
        |n
        Supported output formats: {}|n
        |n
        Examples:|n
        - Download the MNIST dataset:|n
        |s|s%(prog)s -i tfds:mnist -- --save-media|n
        |n
        - Download the VOC 2012 dataset, saving only the annotations in the COCO
          format into a specific directory:|n
        |s|s%(prog)s -i tfds:voc/2012 -f coco -o path/I/like/
        """.format(
            available_datasets, ", ".join(builtin_writers)
        ),
        formatter_class=MultilineFormatter,
    )

    parser.add_argument("-i", "--dataset-id", required=True, help="Which dataset to download")
    parser.add_argument(
        "-f", "--output-format", help="Output format (default: original format of the dataset)"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="Directory to save output (default: a subdir in the current one)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument("-s", "--subset", help="Save only the specified subset")
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for output format (pass '-- -h' for help). "
        "Must be specified after the main command arguments",
    )

    parser.set_defaults(command=download_command)

    return parser


def build_describe_subparser(subparsers: argparse._SubParsersAction):
    parser = subparsers.add_parser(
        name="describe",
        help="Print information about downloadable datasets",
        description="""
        Reports information about datasets that can be downloaded with the
        "datum download" command. The information is reported either as
        human-readable text (the default) or as a JSON object. More detailed
        information can be found in the TFDS Catalog:
        <https://www.tensorflow.org/datasets/catalog/overview>.
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "--report-format",
        choices=("text", "json"),
        default="text",
        help="Format in which to report the information (default: %(default)s)",
    )
    parser.add_argument(
        "--report-file", help="File to which to write the report (default: standard output)"
    )
    parser.set_defaults(command=describe_downloads_command)

    return parser


def get_sensitive_args():
    return {
        download_command: ["dst_dir", "extra_args"],
        describe_downloads_command: ["report-file"],
    }


def download_command(args):
    env = DEFAULT_ENVIRONMENT

    if args.dataset_id.startswith("tfds:"):
        if TFDS_EXTRACTOR_AVAILABLE:
            tfds_ds_name = args.dataset_id[5:]
            tfds_ds = AVAILABLE_TFDS_DATASETS.get(tfds_ds_name)
            if tfds_ds:
                default_output_format = tfds_ds.metadata.default_output_format
                extractor_factory = tfds_ds.make_extractor
            else:
                raise CliException(f"Unsupported TFDS dataset '{tfds_ds_name}'")
        else:
            raise CliException(
                "TFDS datasets are not available, because TFDS and/or "
                "TensorFlow are not installed.\n"
                "You can install them with: pip install datumaro[tf,tfds]"
            )
    else:
        raise CliException(f"Unknown dataset ID '{args.dataset_id}'")

    output_format = args.output_format or default_output_format

    try:
        exporter = env.exporters[output_format]
    except KeyError:
        raise CliException("Exporter for format '%s' is not found" % output_format)
    extra_args = exporter.parse_cmdline(args.extra_args)

    dst_dir = args.dst_dir
    if dst_dir:
        if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
            raise CliException(
                "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
            )
    else:
        dst_dir = generate_next_file_name(
            "%s-%s"
            % (
                make_file_name(args.dataset_id),
                make_file_name(output_format),
            )
        )
    dst_dir = osp.abspath(dst_dir)

    log.info("Downloading the dataset")
    extractor = extractor_factory()

    if args.subset:
        try:
            extractor = extractor.subsets()[args.subset]
        except KeyError:
            raise CliException("Subset '%s' is not present in the dataset" % args.subset)

    log.info("Exporting the dataset")
    exporter.convert(extractor, dst_dir, default_image_ext=".png", **extra_args)

    log.info("Dataset exported to '%s' as '%s'" % (dst_dir, output_format))


def describe_downloads_command(args):
    dataset_metas: Dict[str, TfdsDatasetRemoteMetadata] = {}

    if TFDS_EXTRACTOR_AVAILABLE:
        for dataset_name, dataset in AVAILABLE_TFDS_DATASETS.items():
            dataset_metas[f"tfds:{dataset_name}"] = dataset.query_remote_metadata()

    if args.report_format == "text":
        with open(
            args.report_file, "w"
        ) if args.report_file else contextlib.nullcontext() as report_file:
            if dataset_metas:
                print("Available datasets:", file=report_file)

                for name, meta in sorted(dataset_metas.items()):
                    print(file=report_file)
                    print(f"{name} ({meta.human_name}):", file=report_file)
                    print(
                        f"  default output format: {meta.default_output_format}",
                        file=report_file,
                    )

                    print("  description:", file=report_file)
                    for line in meta.description.rstrip("\n").split("\n"):
                        print(f"    {line}", file=report_file)

                    print(f"  download size: {meta.download_size} bytes", file=report_file)
                    print(f"  home URL: {meta.home_url or 'N/A'}", file=report_file)
                    print(f"  number of classes: {meta.num_classes}", file=report_file)
                    print("  subsets:", file=report_file)
                    for subset_name, subset_meta in sorted(meta.subsets.items()):
                        print(f"    {subset_name}: {subset_meta.num_items} items", file=report_file)
                    print(f"  version: {meta.version}", file=report_file)
            else:
                print("No datasets available.", file=report_file)
                print(file=report_file)
                print(
                    "You can enable TFDS datasets by installing "
                    "TensorFlow and TensorFlow Datasets:",
                    file=report_file,
                )
                print("    pip install datumaro[tf,tfds]", file=report_file)

    elif args.report_format == "json":

        def meta_to_raw(meta: TfdsDatasetRemoteMetadata):
            raw = {}

            # We omit the media type from the output, because there is currently no mechanism
            # for mapping media types to strings. The media type could be useful information
            # for users, though, so we might want to implement such a mechanism eventually.

            for attribute in (
                "default_output_format",
                "description",
                "download_size",
                "home_url",
                "human_name",
                "num_classes",
                "version",
            ):
                raw[attribute] = getattr(meta, attribute)

            raw["subsets"] = {
                name: {"num_items": subset.num_items} for name, subset in meta.subsets.items()
            }

            return raw

        with (
            open(args.report_file, "w") if args.report_file else contextlib.nullcontext(sys.stdout)
        ) as report_file:
            report_file.write(
                dump_json(
                    {name: meta_to_raw(meta) for name, meta in dataset_metas.items()},
                    indent=True,
                    append_newline=True,
                ).decode()
            )
    else:
        assert False, "unreachable code"



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import Optional, Sequence, TypeVar

from datumaro.components.contexts.importer import _ImportFail
from datumaro.components.dataset_base import DatasetBase, SubsetBase

T = TypeVar("T")


def check_identicalness(seq: Sequence[T], raise_error_on_empty: bool = True) -> Optional[T]:
    if len(seq) == 0 and raise_error_on_empty:
        raise _ImportFail("It should not be empty.")
    elif len(seq) == 0 and not raise_error_on_empty:
        return None

    if seq.count(seq[0]) != len(seq):
        raise _ImportFail("All items in the sequence should be identical.")

    return seq[0]


class ExtractorMerger(DatasetBase):
    """A simple class to merge single-subset extractors."""

    def __init__(
        self,
        sources: Sequence[SubsetBase],
    ):
        if len(sources) == 0:
            raise _ImportFail("It should not be empty.")

        self._infos = check_identicalness([s.infos() for s in sources])
        self._categories = check_identicalness([s.categories() for s in sources])
        self._media_type = check_identicalness([s.media_type() for s in sources])
        self._is_stream = check_identicalness([s.is_stream for s in sources])
        self._subsets = {s.subset: s for s in sources}

    def infos(self):
        return self._infos

    def categories(self):
        return self._categories

    def __iter__(self):
        for subset in self._subsets.values():
            yield from subset

    def __len__(self):
        return sum(len(subset) for subset in self._subsets.values())

    def get(self, id: str, subset: Optional[str] = None):
        if subset is None:
            for s in self._subsets.values():
                item = s.get(id)
                if item is not None:
                    return item

        s = self._subsets[subset]
        return s.get(id, subset)

    @property
    def is_stream(self) -> bool:
        return self._is_stream



# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import os.path as osp
from collections import OrderedDict
from glob import glob
from typing import Optional

import numpy as np

from datumaro.components.annotation import AnnotationType, LabelCategories, Mask, MaskCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import InvalidAnnotationError
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image
from datumaro.util.image import find_images
from datumaro.util.mask_tools import generate_colormap, load_mask
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file

from .format import (
    SynthiaAlLabelMap,
    SynthiaAlPath,
    SynthiaRandLabelMap,
    SynthiaRandPath,
    SynthiaSfLabelMap,
    SynthiaSfPath,
)


def make_categories(label_map):
    categories = {}
    label_categories = LabelCategories()
    for label in label_map:
        label_categories.add(label)
    categories[AnnotationType.label] = label_categories

    has_colors = any(v is not None for v in label_map.values())
    if not has_colors:  # generate new colors
        colormap = generate_colormap(len(label_map))
    else:  # only copy defined colors
        colormap = {
            label_id: (desc[0], desc[1], desc[2])
            for label_id, desc in enumerate(label_map.values())
        }
    mask_categories = MaskCategories(colormap)
    mask_categories.inverse_colormap  # pylint: disable=pointless-statement
    categories[AnnotationType.mask] = mask_categories
    return categories


def parse_label_map(path):
    label_map = OrderedDict()
    with open(path, "r", encoding="utf-8") as f:
        for line in f:
            # skip empty and commented lines
            line = line.strip()
            if not line or line[0] == "#":
                continue

            # color, name
            label_desc = line.split()

            if 2 < len(label_desc):
                name = label_desc[3]
                color = tuple([int(c) for c in label_desc[:3]])
            else:
                name = label_desc[0]
                color = None

            if name in label_map:
                raise InvalidAnnotationError("Label '%s' is already defined" % name)

            label_map[name] = color
    return label_map


class _SynthiaBase(SubsetBase):
    def __init__(
        self,
        path: str,
        path_formats,
        label_map,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isdir(path):
            raise NotADirectoryError(errno.ENOTDIR, "Can't find dataset directory", path)

        super().__init__(subset=subset, ctx=ctx)

        self._path_formats = path_formats
        self._label_map = label_map

        self._img_dir = None
        self._inst_dir = None
        self._seg_dir = None
        for path_format in vars(path_formats).keys():
            if path_format == "IMAGES_DIR":
                self._img_dir = osp.join(path, path_formats.IMAGES_DIR)
            elif path_format == "LABELS_SEGM_DIR":
                self._inst_dir = osp.join(path, path_formats.LABELS_SEGM_DIR)
            elif path_format == "SEMANTIC_SEGM_DIR":
                self._seg_dir = osp.join(path, path_formats.SEMANTIC_SEGM_DIR)

        self._categories = self._load_categories(path)
        self._items = list(self._load_items().values())

    def _load_categories(self, path):
        if has_meta_file(path):
            return make_categories(parse_meta_file(path))

        label_map_path = osp.join(path, "label_colors.txt")
        if osp.isfile(label_map_path):
            label_map = parse_label_map(label_map_path)
        else:
            label_map = self._label_map

        return make_categories(label_map)

    def _load_items(self):
        if self._img_dir and osp.isdir(self._img_dir):
            images = {
                osp.splitext(osp.relpath(p, self._img_dir))[0].replace("\\", "/"): p
                for p in find_images(self._img_dir, recursive=True)
            }
        else:
            images = {}

        items = {}
        if self._inst_dir and osp.isdir(self._inst_dir):
            gt_labels = glob(self._inst_dir + "/*.txt")
            for gt_label in gt_labels:
                item_id = osp.splitext(osp.relpath(gt_label, self._inst_dir))[0].replace("\\", "/")

                anno = []
                labels_mask = np.loadtxt(gt_label)
                classes = np.unique(labels_mask)
                for label_id in classes:
                    anno.append(
                        Mask(image=self._lazy_extract_mask(labels_mask, label_id), label=label_id)
                    )

                image = images.get(item_id)
                if image:
                    image = Image.from_file(path=image)

                items[item_id] = DatasetItem(id=item_id, media=image, annotations=anno)
        elif self._seg_dir and osp.isdir(self._seg_dir):
            for seg_img_path in find_images(self._seg_dir, recursive=True):
                item_id = osp.splitext(osp.relpath(seg_img_path, self._seg_dir))[0].replace(
                    "\\", "/"
                )

                inverse_cls_colormap = self._categories[AnnotationType.mask].inverse_colormap

                color_mask = load_mask(seg_img_path, inverse_cls_colormap, default_id=0)

                anno = []
                classes = np.unique(color_mask)
                for label_id in classes:
                    anno.append(
                        Mask(image=self._lazy_extract_mask(color_mask, label_id), label=label_id)
                    )

                image = images.get(item_id)
                if image:
                    image = Image.from_file(path=image)

                items[item_id] = DatasetItem(id=item_id, media=image, annotations=anno)

        return items

    @staticmethod
    def _lazy_extract_mask(mask, c):
        return lambda: mask == c


class SynthiaRandBase(_SynthiaBase):
    def __init__(self, path: str, **kwargs):
        super().__init__(
            path=path, path_formats=SynthiaRandPath, label_map=SynthiaRandLabelMap, **kwargs
        )


class SynthiaSfBase(_SynthiaBase):
    def __init__(self, path, **kwargs):
        super().__init__(
            path=path, path_formats=SynthiaSfPath, label_map=SynthiaSfLabelMap, **kwargs
        )


class SynthiaAlBase(_SynthiaBase):
    def __init__(self, path, **kwargs):
        super().__init__(
            path=path, path_formats=SynthiaAlPath, label_map=SynthiaAlLabelMap, **kwargs
        )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import base64

import pyarrow as pa


def pa_batches_decoder(batches, column=None):
    if not isinstance(batches, list):
        batches = [batches]
    table = pa.Table.from_batches(batches)
    if column:
        data = table.column(column).to_pylist()
    else:
        data = table.to_pylist()
    return data


def b64encode(obj, prefix=None):
    if isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = b64encode(v, prefix)
    elif isinstance(obj, (list, tuple)):
        _obj = []
        for v in obj:
            _obj.append(b64encode(v, prefix))
        if isinstance(obj, list):
            _obj = list(_obj)
        obj = _obj
    elif isinstance(obj, bytes):
        obj = base64.b64encode(obj).decode()
        if prefix:
            obj = prefix + obj
    return obj


def b64decode(obj, prefix=None):
    if isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = b64decode(v, prefix)
    elif isinstance(obj, (list, tuple)):
        _obj = []
        for v in obj:
            _obj.append(b64decode(v, prefix))
        if isinstance(obj, list):
            _obj = list(_obj)
        obj = _obj
    elif isinstance(obj, str):
        if prefix and obj.startswith(prefix):
            obj = obj.replace(prefix, "", 1)
            obj = base64.b64decode(obj)
        else:
            try:
                _obj = base64.b64decode(obj)
                if base64.b64encode(_obj).decode() == obj:
                    obj = _obj
            except Exception:
                pass
    return obj



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT


import pyarrow as pa

from datumaro.errors import DatasetImportError


class DatumaroArrow:
    SIGNATURE = "signature:datumaro_arrow"
    VERSION = "1.0"

    MP_TIMEOUT = 300.0  # 5 minutes

    SCHEMA = pa.schema(
        [
            pa.field("id", pa.string()),
            pa.field("subset", pa.string()),
            pa.field(
                "media",
                pa.struct(
                    [
                        pa.field("type", pa.uint32()),
                        pa.field("path", pa.string()),
                        pa.field("bytes", pa.binary()),
                        pa.field("attributes", pa.binary()),
                    ]
                ),
            ),
            pa.field("annotations", pa.binary()),
            pa.field("attributes", pa.binary()),
        ]
    )

    @classmethod
    def check_signature(cls, signature: str):
        if signature != cls.SIGNATURE:
            raise DatasetImportError(
                f"Input signature={signature} is not aligned with the ground truth signature={cls.SIGNATURE}"
            )

    @classmethod
    def check_schema(cls, schema: pa.lib.Schema):
        if not cls.SCHEMA.equals(schema):
            raise DatasetImportError(
                f"input schema is not aligned with the ground truth schema.\n"
                f"input schema: \n{schema.to_string()}\n"
                f"ground truth schema: \n{cls.SCHEMA.to_string()}\n"
            )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
import re
from typing import Dict, List, Optional

import pyarrow as pa

from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer

from .format import DatumaroArrow


class ArrowImporter(Importer):
    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> Optional[FormatDetectionConfidence]:
        def verify_datumaro_arrow_format(file):
            with pa.ipc.open_stream(file) as reader:
                schema = reader.schema
            DatumaroArrow.check_signature(schema.metadata.get(b"signature", b"").decode())
            DatumaroArrow.check_schema(schema)

        if context.root_path.endswith(".arrow"):
            verify_datumaro_arrow_format(context.root_path)
        else:
            for arrow_file in context.require_files("*.arrow"):
                with context.probe_text_file(
                    arrow_file,
                    f"{arrow_file} is not Datumaro arrow format.",
                    is_binary_file=True,
                ) as f:
                    verify_datumaro_arrow_format(f)

    @classmethod
    def find_sources(cls, path) -> List[Dict]:
        sources = cls._find_sources_recursive(
            path,
            ".arrow",
            cls.NAME,
        )

        # handle sharded arrow files
        _sources = []
        for source in sources:
            match = re.match(r"(.*?)(-[0-9]+-of-[0-9]+)?\.arrow", source["url"])
            found = False
            prefix = match.group(1)
            for _source in _sources:
                if _source["url"].startswith(prefix):
                    _source["options"]["additional_paths"].append(source["url"])
                    found = True
                    break
            if not found:
                source["options"] = {
                    "additional_paths": [],
                    "subset": os.path.basename(prefix),
                }
                _sources.append(source)
        return _sources



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT


import os
import struct
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pyarrow as pa

from datumaro.components.errors import DatumaroError
from datumaro.components.media import Image, MediaElement, MediaType, PointCloud
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import DictMapper, Mapper
from datumaro.util.image import decode_image, encode_image, load_image

from .utils import b64decode, b64encode, pa_batches_decoder


class MediaMapper(Mapper):
    @classmethod
    def forward(cls, obj: Optional[MediaElement], **options) -> Dict[str, Any]:
        if obj is None:
            return {"type": int(MediaType.NONE)}
        if obj._type == MediaType.IMAGE:
            return ImageMapper.forward(obj, **options)
        if obj._type == MediaType.POINT_CLOUD:
            return PointCloudMapper.forward(obj, **options)
        if obj._type == MediaType.MEDIA_ELEMENT:
            return MediaElementMapper.forward(obj)
        raise DatumaroError(f"{obj._type} is not allowed for MediaMapper.")

    @classmethod
    def backward(cls, obj: Dict[str, Any]) -> Optional[MediaElement]:
        media_type = obj["type"]

        if media_type == MediaType.NONE:
            return None
        if media_type == MediaType.IMAGE:
            return ImageMapper.backward(obj)
        if media_type == MediaType.POINT_CLOUD:
            return PointCloudMapper.backward(obj)
        if media_type == MediaType.MEDIA_ELEMENT:
            return MediaElementMapper.backward(obj)
        raise DatumaroError(f"{media_type} is not allowed for MediaMapper.")

    @classmethod
    def backward_from_batches(
        cls,
        batches: List[pa.lib.RecordBatch],
        parent: Optional[str] = None,
    ) -> List[Optional[MediaElement]]:
        types = pa_batches_decoder(batches, f"{parent}.type" if parent else "type")
        assert len(set(types)) == 1, "The types in batch are not identical."

        if types[0] == MediaType.NONE:
            return [None for _ in types]
        if types[0] == MediaType.IMAGE:
            return ImageMapper.backward_from_batches(batches, parent)
        if types[0] == MediaType.POINT_CLOUD:
            return PointCloudMapper.backward_from_batches(batches, parent)
        if types[0] == MediaType.MEDIA_ELEMENT:
            return MediaElementMapper.backward_from_batches(batches, parent)
        raise NotImplementedError


class MediaElementMapper(Mapper):
    MAGIC_PATH = "/NOT/A/REAL/PATH"
    MEDIA_TYPE = MediaType.MEDIA_ELEMENT

    @classmethod
    def forward(cls, obj: MediaElement) -> Dict[str, Any]:
        return {
            "type": int(obj.type),
            "path": getattr(obj, "path", cls.MAGIC_PATH),
        }

    @classmethod
    def backward_dict(cls, obj: Dict[str, Any]) -> Dict[str, Any]:
        obj = obj.copy()
        media_type = obj["type"]
        assert media_type == cls.MEDIA_TYPE, f"Expect {cls.MEDIA_TYPE} but actual is {media_type}."
        obj["type"] = media_type
        return obj

    @classmethod
    def backward(cls, obj: Dict[str, Any]) -> MediaElement:
        _ = cls.backward_dict(obj)
        return MediaElement()

    @classmethod
    def backward_from_batches(
        cls,
        batches: List[pa.lib.RecordBatch],
        parent: Optional[str] = None,
    ) -> List[MediaElement]:
        types = pa_batches_decoder(batches, f"{parent}.type" if parent else "type")
        return [MediaElement() for _ in types]


class ImageMapper(MediaElementMapper):
    MEDIA_TYPE = MediaType.IMAGE
    AVAILABLE_SCHEMES = ("AS-IS", "PNG", "TIFF", "JPEG/95", "JPEG/75", "NONE")

    @classmethod
    def encode(cls, obj: Image, scheme: str = "PNG") -> Optional[bytes]:
        if scheme is None or scheme == "NONE":
            return None
        if scheme == "AS-IS":
            _bytes = obj.bytes
            if _bytes is not None:
                return _bytes
            # try to encode in PNG
            scheme = "PNG"

        options = {}
        if scheme.startswith("JPEG"):
            quality = int(scheme.split("/")[-1])
            options["ext"] = "JPEG"
            options["jpeg_quality"] = quality
        elif scheme == "PNG":
            options["ext"] = "PNG"
        elif scheme == "TIFF":
            options["ext"] = "TIFF"
        else:
            raise NotImplementedError

        data = obj.data
        if data is not None:
            return encode_image(obj.data, **options)
        return None

    @classmethod
    def decode(
        cls, path: Optional[str] = None, data: Optional[bytes] = None
    ) -> Optional[np.ndarray]:
        if path is None and data is None:
            return None
        if data is not None:
            return decode_image(data, np.uint8)
        if path is not None:
            return load_image(path, np.uint8)

    @classmethod
    def forward(
        cls, obj: Image, encoder: Union[str, Callable[[Image], bytes]] = "PNG"
    ) -> Dict[str, Any]:
        out = super().forward(obj)

        _bytes = None
        if isinstance(encoder, Callable):
            _bytes = encoder(obj)
        else:
            _bytes = cls.encode(obj, scheme=encoder)
        out["bytes"] = _bytes

        out["attributes"] = DictMapper.forward(dict(size=obj.size))

        return out

    @classmethod
    def backward(cls, obj: Dict[str, Any]) -> Image:
        media_dict = cls.backward_dict(obj)

        path = media_dict["path"]
        attributes, _ = DictMapper.backward(media_dict["attributes"], 0)

        _bytes = None
        if media_dict.get("bytes", None) is not None:
            _bytes = media_dict["bytes"]

        if _bytes:
            return Image.from_bytes(data=_bytes, size=attributes["size"])
        return Image.from_file(path=path, size=attributes["size"])

    @classmethod
    def backward_from_batches(
        cls,
        batches: List[pa.lib.RecordBatch],
        parent: Optional[str] = None,
    ) -> List[Image]:
        paths = pa_batches_decoder(batches, f"{parent}.path" if parent else "path")
        attributes_ = pa_batches_decoder(
            batches, f"{parent}.attributes" if parent else "attributes"
        )
        attributes_ = [DictMapper.backward(attributes)[0] for attributes in attributes_]

        images = []
        for idx, (path, attributes) in enumerate(zip(paths, attributes_)):
            if os.path.exists(path):
                images.append(Image.from_file(path=path, size=attributes["size"]))
            else:
                images.append(
                    Image.from_bytes(
                        data=lambda: pa_batches_decoder(
                            batches, f"{parent}.bytes" if parent else "bytes"
                        )[idx],
                        size=attributes["size"],
                    )
                )
        return images


# TODO: share binary for extra images
class PointCloudMapper(MediaElementMapper):
    MEDIA_TYPE = MediaType.POINT_CLOUD
    B64_PREFIX = "//B64_ENCODED//"

    @classmethod
    def forward(
        cls, obj: PointCloud, encoder: Union[str, Callable[[PointCloud], bytes]] = "PNG"
    ) -> Dict[str, Any]:
        out = super().forward(obj)

        _bytes = None
        if isinstance(encoder, Callable):
            _bytes = encoder(obj)
        elif encoder != "NONE":
            _bytes = obj.data
        out["bytes"] = _bytes

        bytes_arr = bytearray()
        bytes_arr.extend(struct.pack("<I", len(obj.extra_images)))
        for img in obj.extra_images:
            bytes_arr.extend(
                DictMapper.forward(
                    b64encode(ImageMapper.forward(img, encoder=encoder), cls.B64_PREFIX)
                )
            )
        out["attributes"] = bytes(bytes_arr)

        return out

    @classmethod
    def backward(cls, obj: Dict[str, Any]) -> PointCloud:
        offset = 0
        media_dict = cls.backward_dict(obj)

        path = media_dict["path"]
        (len_extra_images,) = struct.unpack_from("<I", media_dict["attributes"], offset)
        offset += 4

        _bytes = None
        if media_dict.get("bytes", None) is not None:
            _bytes = media_dict["bytes"]

        extra_images = []
        for _ in range(len_extra_images):
            img, offset = DictMapper.backward(media_dict["attributes"], offset)
            extra_images.append(ImageMapper.backward(b64decode(img, cls.B64_PREFIX)))

        if _bytes:
            return PointCloud.from_bytes(data=_bytes, extra_images=extra_images)
        else:
            return PointCloud.from_file(path=path, extra_images=extra_images)

    @classmethod
    def backward_from_batches(
        cls,
        batches: List[pa.lib.RecordBatch],
        parent: Optional[str] = None,
    ) -> List[PointCloud]:
        paths = pa_batches_decoder(batches, f"{parent}.path" if parent else "path")

        def extra_images(idx):
            offset = 0
            attributes = pa_batches_decoder(
                batches, f"{parent}.attributes" if parent else "attributes"
            )[idx]
            (len_extra_images,) = struct.unpack_from("<I", attributes, offset)
            offset += 4
            outs = []
            for _ in range(len_extra_images):
                img, offset = DictMapper.backward(attributes, offset)
                outs.append(ImageMapper.backward(b64decode(img, cls.B64_PREFIX)))
            return outs

        point_clouds = []
        for idx, path in enumerate(paths):
            if os.path.exists(path):
                point_clouds.append(
                    PointCloud.from_file(
                        path=path,
                        extra_images=partial(extra_images, idx=idx),
                    )
                )
            else:
                point_clouds.append(
                    PointCloud.from_bytes(
                        data=lambda: pa_batches_decoder(
                            batches, f"{parent}.bytes" if parent else "bytes"
                        )[idx],
                        extra_images=partial(extra_images, idx=idx),
                    )
                )
        return point_clouds



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import glob
import importlib
import logging as log
import os.path as osp
from functools import partial
from inspect import isclass
from typing import (
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
)

from datumaro.components.cli_plugin import CliPlugin, plugin_types
from datumaro.components.format_detection import (
    DetectedFormat,
    FormatDetectionConfidence,
    RejectionReason,
    detect_dataset_format,
)
from datumaro.components.lazy_plugin import PLUGIN_TYPES, LazyPlugin
from datumaro.util.os_util import import_foreign_module, split_path

T = TypeVar("T")


class Registry(Generic[T]):
    def __init__(self):
        self._items: Dict[str, T] = {}

    def register(self, name: str, value: T) -> T:
        self._items[name] = value
        return value

    def unregister(self, name: str) -> Optional[T]:
        return self._items.pop(name, None)

    def get(self, key: str):
        """Returns a class or a factory function"""
        return self._items[key]

    def __getitem__(self, key: str) -> T:
        return self.get(key)

    def __contains__(self, key) -> bool:
        return key in self._items

    def __iter__(self) -> Iterator[str]:
        return iter(self._items)

    def items(self) -> Generator[Tuple[str, T], None, None]:
        for key in self:
            yield key, self.get(key)


class PluginRegistry(Registry[Type[CliPlugin]]):
    def __init__(
        self, filter: Callable[[Type[CliPlugin]], bool] = None
    ):  # pylint: disable=redefined-builtin
        super().__init__()
        self._filter = filter

    def get(self, key: str) -> PLUGIN_TYPES:
        """Returns a class or a factory function"""
        item = self._items[key]
        if issubclass(item, LazyPlugin):
            return item.get_plugin_cls()
        return item

    def batch_register(self, values: Iterable[CliPlugin]):
        for v in values:
            if self._filter and not self._filter(v):
                continue

            self.register(v.NAME, v)


class Environment:
    _builtin_plugins = None

    @classmethod
    def _make_filter(cls, accept, decline=None, skip=None):
        accept = (accept,) if isclass(accept) else tuple(accept)
        skip = {skip} if isclass(skip) else set(skip or [])
        skip = tuple(skip | set(accept))
        return partial(cls._check_type, accept=accept, decline=decline, skip=skip)

    @staticmethod
    def _check_type(t, *, accept, decline, skip):
        if not issubclass(t, accept) or t in skip or (decline and issubclass(t, decline)):
            return False
        if getattr(t, "__not_plugin__", None):
            return False
        return True

    def __init__(self, use_lazy_import: bool = True):
        from datumaro.components.dataset_base import DatasetBase, SubsetBase
        from datumaro.components.exporter import Exporter
        from datumaro.components.generator import DatasetGenerator
        from datumaro.components.importer import Importer
        from datumaro.components.launcher import Launcher
        from datumaro.components.transformer import ItemTransform, Transform
        from datumaro.components.validator import Validator

        _filter = self._make_filter
        self._extractors = PluginRegistry(
            _filter(
                DatasetBase,
                decline=Transform,
                skip=(SubsetBase, Transform, ItemTransform),
            )
        )
        self._importers = PluginRegistry(_filter(Importer))
        self._launchers = PluginRegistry(_filter(Launcher))
        self._exporters = PluginRegistry(_filter(Exporter))
        self._generators = PluginRegistry(_filter(DatasetGenerator))
        self._transforms = PluginRegistry(_filter(Transform, skip=ItemTransform))
        self._validators = PluginRegistry(_filter(Validator))
        self._builtins_initialized = False
        self._use_lazy_import = use_lazy_import

    def _get_plugin_registry(self, name):
        if not self._builtins_initialized:
            self._builtins_initialized = True
            self._register_builtin_plugins()
        return getattr(self, name)

    @property
    def extractors(self) -> PluginRegistry:
        return self._get_plugin_registry("_extractors")

    @property
    def importers(self) -> PluginRegistry:
        return self._get_plugin_registry("_importers")

    @property
    def launchers(self) -> PluginRegistry:
        return self._get_plugin_registry("_launchers")

    @property
    def exporters(self) -> PluginRegistry:
        return self._get_plugin_registry("_exporters")

    @property
    def generators(self) -> PluginRegistry:
        return self._get_plugin_registry("_generators")

    @property
    def transforms(self) -> PluginRegistry:
        return self._get_plugin_registry("_transforms")

    @property
    def validators(self) -> PluginRegistry:
        return self._get_plugin_registry("_validators")

    @staticmethod
    def _find_plugins(plugins_dir):
        plugins = []

        for pattern in ("*.py", "*/*.py", "*/*/*.py"):
            for path in glob.glob(osp.join(glob.escape(plugins_dir), pattern)):
                if not osp.isfile(path):
                    continue

                path_rel = osp.relpath(path, plugins_dir)
                name_parts = split_path(osp.splitext(path_rel)[0])

                # a module with a dot in the name won't load correctly
                if any("." in part for part in name_parts):
                    log.warning(
                        "Python file '%s' in directory '%s' can't be imported "
                        "due to a dot in the name; skipping.",
                        path_rel,
                        plugins_dir,
                    )
                    continue
                plugins.append(".".join(name_parts))

        return plugins

    @classmethod
    def _get_plugin_exports(cls, module, types):
        exports = []
        if hasattr(module, "exports"):
            exports = module.exports
        else:
            for symbol in dir(module):
                if symbol.startswith("_"):
                    continue
                exports.append(getattr(module, symbol))

        exports = [s for s in exports if isclass(s) and issubclass(s, types) and s not in types]

        return exports

    @classmethod
    def _load_plugins(cls, module_names, *, importer, types=None):
        types = tuple(types or plugin_types())

        all_exports = []
        for module_name in module_names:
            try:
                module = importer(module_name)
                exports = cls._get_plugin_exports(module, types)
            except Exception as e:
                module_search_error = ModuleNotFoundError

                message = ["Failed to import module '%s': %s", module_name, e]
                if isinstance(e, module_search_error):
                    log.debug(*message)
                else:
                    log.warning(*message)
                continue

            log.debug(
                "Imported the following symbols from %s: %s"
                % (module_name, ", ".join(s.__name__ for s in exports))
            )
            all_exports.extend(exports)

        return all_exports

    @classmethod
    def _load_builtin_plugins(cls):
        """Load builtin plugins which will be imported lazily using plugin spec files"""
        if cls._builtin_plugins is None:
            from datumaro.plugins.specs import get_lazy_plugins

            cls._builtin_plugins = get_lazy_plugins()
        return cls._builtin_plugins

    @classmethod
    def _load_builtin_plugins_from_importlib(cls):
        """Load builtin plugins from importlib, not lazy import from plugin spec files"""
        if cls._builtin_plugins is None:
            import datumaro.plugins

            plugins_dir = osp.dirname(datumaro.plugins.__file__)
            module_names = [
                datumaro.plugins.__name__ + "." + name for name in cls._find_plugins(plugins_dir)
            ]
            cls._builtin_plugins = cls._load_plugins(module_names, importer=importlib.import_module)
        return cls._builtin_plugins

    def load_plugins(self, plugins_dir):
        module_names = self._find_plugins(plugins_dir)
        plugins = self._load_plugins(
            module_names, importer=partial(import_foreign_module, path=plugins_dir)
        )

        self.register_plugins(plugins)

    def _register_builtin_plugins(self):
        self.register_plugins(
            self._load_builtin_plugins()
            if self._use_lazy_import
            else self._load_builtin_plugins_from_importlib()
        )

    def register_plugins(self, plugins):
        self.extractors.batch_register(plugins)
        self.importers.batch_register(plugins)
        self.launchers.batch_register(plugins)
        self.exporters.batch_register(plugins)
        self.generators.batch_register(plugins)
        self.transforms.batch_register(plugins)
        self.validators.batch_register(plugins)

    def make_extractor(self, name, *args, **kwargs):
        return self.extractors.get(name)(*args, **kwargs)

    def make_importer(self, name, *args, **kwargs):
        return self.importers.get(name)(*args, **kwargs)

    def make_launcher(self, name, *args, **kwargs):
        return self.launchers.get(name)(*args, **kwargs)

    def make_exporter(self, name, *args, **kwargs):
        result = self.exporters.get(name)
        if isclass(result):
            result = result.convert
        return partial(result, *args, **kwargs)

    def make_transform(self, name, *args, **kwargs):
        return partial(self.transforms.get(name), *args, **kwargs)

    def is_format_known(self, name):
        return name in self.importers or name in self.extractors

    def detect_dataset(
        self,
        path: str,
        depth: int = 1,
        rejection_callback: Optional[Callable[[str, RejectionReason, str], None]] = None,
    ) -> List[str]:
        ignore_dirs = {"__MSOSX", "__MACOSX"}
        all_matched_formats: Set[DetectedFormat] = set()

        for _ in range(depth + 1):
            detected_formats = detect_dataset_format(
                (
                    (format_name, importer.detect)
                    for format_name, importer in self.importers.items()
                ),
                path,
                rejection_callback=rejection_callback,
            )

            if detected_formats:
                all_matched_formats |= set(detected_formats)

            paths = glob.glob(osp.join(path, "*"))
            path = "" if len(paths) != 1 else paths[0]
            if not osp.isdir(path) or osp.basename(path) in ignore_dirs:
                break

        max_conf = (
            max(all_matched_formats).confidence
            if len(all_matched_formats) > 0
            else FormatDetectionConfidence.NONE
        )

        return sorted(
            [str(format) for format in all_matched_formats if format.confidence == max_conf]
        )

    def __reduce__(self):
        return (self.__class__, ())

    @classmethod
    def merge(cls, envs: Sequence["Environment"]) -> "Environment":
        if all([env == DEFAULT_ENVIRONMENT for env in envs]):
            return DEFAULT_ENVIRONMENT

        merged = Environment()

        def _register(registry: PluginRegistry):
            merged.register_plugins(plugin for plugin in registry)

        for env in envs:
            _register(env.extractors)
            _register(env.importers)
            _register(env.launchers)
            _register(env.exporters)
            _register(env.generators)
            _register(env.transforms)
            _register(env.validators)

        return merged

    @classmethod
    def release_builtin_plugins(cls):
        cls._builtin_plugins = None


DEFAULT_ENVIRONMENT = Environment()



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import inspect
import os
import os.path as osp
import shutil
from typing import Dict, Iterable, Optional, Type, Union

from datumaro.cli.util.compare import DistanceCompareVisualizer
from datumaro.cli.util.project import generate_next_file_name
from datumaro.components.comparator import DistanceComparator, EqualityComparator, TableComparator
from datumaro.components.dataset import Dataset, DatasetItemStorageDatasetView, IDataset
from datumaro.components.environment import Environment
from datumaro.components.errors import DatasetError
from datumaro.components.exporter import Exporter
from datumaro.components.filter import XPathAnnotationsFilter, XPathDatasetFilter
from datumaro.components.launcher import Launcher
from datumaro.components.merge import DEFAULT_MERGE_POLICY, get_merger
from datumaro.components.transformer import ModelTransform, Transform
from datumaro.components.validator import TaskType, Validator
from datumaro.util import dump_json_file, parse_str_enum_value
from datumaro.util.scope import on_error_do, scoped

__all__ = ["HLOps"]


class HLOps:
    """High-level dataset operations for Python API."""

    @staticmethod
    def compare(
        first_dataset: IDataset,
        second_dataset: IDataset,
        report_dir: Optional[str] = None,
        method: str = "table",
        **kwargs,
    ) -> IDataset:
        """
        Compare two datasets and optionally save a comparison report.

        Args:
            first_dataset (IDataset): The first dataset to compare.
            second_dataset (IDataset): The second dataset to compare.
            report_dir (Optional[str], optional): The directory path to save the comparison report. Defaults to None.
            method (str, optional): The comparison method to use. Possible values are "table", "equality", "distance".
                Defaults to "table".
            **kwargs: Additional keyword arguments that can be passed to the comparison method.

        Returns:
            IDataset: The result of the comparison.

        Raises:
            ValueError: If the method is "distance" and report_dir is not specified.

        Example:
            comparator = Comparator()
            result = comparator.compare(first_dataset, second_dataset, report_dir="./comparison_report")
            print(result)
        """
        if method == "table":
            comparator = TableComparator()
            h_table, m_table, l_table, result_dict = comparator.compare_datasets(
                first_dataset, second_dataset
            )
            if report_dir:
                comparator.save_compare_report(h_table, m_table, l_table, result_dict, report_dir)

        elif method == "equality":
            comparator = EqualityComparator(**kwargs)
            output = comparator.compare_datasets(first_dataset, second_dataset)
            if report_dir:
                comparator.save_compare_report(output, report_dir)

        elif method == "distance":
            if not report_dir:
                raise ValueError(
                    "Please specify report_dir to save comparision result for DistanceComparator."
                )
            output_format = kwargs.pop("output_format", "simple")
            comparator = DistanceComparator(**kwargs)
            with DistanceCompareVisualizer(
                save_dir=report_dir,
                comparator=comparator,
                output_format=output_format,
            ) as visualizer:
                visualizer.save(first_dataset, second_dataset)

        return 0

    @staticmethod
    def transform(
        dataset: IDataset,
        method: Union[str, Type[Transform]],
        *,
        env: Optional[Environment] = None,
        **kwargs,
    ) -> IDataset:
        """
        Applies some function to dataset items.

        Results are computed lazily, if the transform supports this.

        Args:
            dataset: The dataset to be transformed
            method: The transformation to be applied to the dataset.
                If a string is passed, it is treated as a plugin name,
                which is searched for in the environment
                set by the 'env' argument
            env: A plugin collection. If not set, the built-in plugins are used
            **kwargs: Parameters for the transformation

        Returns: a wrapper around the input dataset
        """

        if isinstance(method, str):
            if env is None:
                env = Environment()
            method = env.transforms[method]

        if not (inspect.isclass(method) and issubclass(method, Transform)):
            raise TypeError(f"Unexpected 'method' argument type: {type(method)}")

        produced = method(dataset, **kwargs)

        return Dataset(source=produced, env=env)

    @staticmethod
    def filter(
        dataset: IDataset,
        expr: str,
        *,  # pylint: disable=redefined-builtin
        filter_annotations: bool = False,
        remove_empty: bool = False,
    ) -> IDataset:
        """
        Filters out some dataset items or annotations, using a custom filter
        expression.

        Args:
            dataset: The dataset to be filtered
            expr: XPath-formatted filter expression
                (e.g. `/item[subset = 'train']`, `/item/annotation[label = 'cat']`)
            filter_annotations: Indicates if the filter should be
                applied to items or annotations
            remove_empty: When filtering annotations, allows to
                exclude empty items from the resulting dataset

        Returns: a wrapper around the input dataset, which is computed lazily
            during iteration
        """

        if filter_annotations:
            return HLOps.transform(
                dataset, XPathAnnotationsFilter, xpath=expr, remove_empty=remove_empty
            )
        else:
            if not expr:
                return dataset
            return HLOps.transform(dataset, XPathDatasetFilter, xpath=expr)

    @staticmethod
    def merge(
        *datasets: Dataset,
        merge_policy: str = DEFAULT_MERGE_POLICY,
        report_path: Optional[str] = None,
        **kwargs,
    ) -> Dataset:
        """
        Merge `datasets` according to `merge_policy`. You have to choose an appropriate `merge_policy`
        for your purpose. The available merge policies are "union", "intersect", and "exact".
        For more details about the merge policies, please refer to :func:`get_merger`.
        """

        merger = get_merger(merge_policy, **kwargs)
        merged = merger(*datasets)
        env = Environment.merge(
            dataset.env
            for dataset in datasets
            if hasattr(
                dataset, "env"
            )  # TODO: Sometimes, there is dataset which is not exactly "Dataset",
            # e.g., VocClassificationBase. this should be fixed and every object from
            # Dataset.import_from should have "Dataset" type.
        )
        if report_path:
            merger.save_merge_report(report_path)
        return Dataset(source=merged, env=env)

    @staticmethod
    def run_model(
        dataset: IDataset,
        model: Union[Launcher, Type[ModelTransform]],
        *,
        batch_size: int = 1,
        append_annotation: bool = False,
        num_workers: int = 0,
        **kwargs,
    ) -> IDataset:
        """
        Run the model on the dataset item media entities, such as images,
        to obtain pseudo labels and add them as dataset annotations.

        Args:
            dataset: The dataset to be transformed
            model: The model to be applied to the dataset
            batch_size: The number of dataset items processed
                simultaneously by the model
            append_annotation: Whether append new annotation to existed annotations
            num_workers: The number of worker threads to use for parallel inference.
                Set to 0 for single-process mode. Default is 0.
            **kwargs: Parameters for the model

        Returns: a wrapper around the input dataset, which is computed lazily
            during iteration
        """

        if isinstance(model, Launcher):
            return HLOps.transform(
                dataset,
                ModelTransform,
                launcher=model,
                batch_size=batch_size,
                append_annotation=append_annotation,
                num_workers=num_workers,
                **kwargs,
            )
        elif inspect.isclass(model) and issubclass(model, ModelTransform):
            return HLOps.transform(
                dataset,
                model,
                batch_size=batch_size,
                append_annotation=append_annotation,
                num_workers=num_workers,
                **kwargs,
            )
        else:
            raise TypeError(f"Unexpected model argument type: {type(model)}")

    @staticmethod
    @scoped
    def export(
        dataset: IDataset,
        path: str,
        format: Union[str, Type[Exporter]],
        *,
        env: Optional[Environment] = None,
        **kwargs,
    ) -> None:
        """
        Saves the input dataset in some format.

        Args:
            dataset: The dataset to be saved
            path: The output directory
            format: The desired output format for the dataset.
                If a string is passed, it is treated as a plugin name,
                which is searched for in the environment set by the 'env' argument
            env: A plugin collection. If not set, the built-in plugins are used
            **kwargs: Parameters for the export format
        """

        if isinstance(format, str):
            if env is None:
                env = Environment()
            exporter = env.exporters[format]
        else:
            exporter = format

        if not (inspect.isclass(exporter) and issubclass(exporter, Exporter)):
            raise TypeError(f"Unexpected 'format' argument type: {type(exporter)}")

        path = osp.abspath(path)
        if not osp.exists(path):
            on_error_do(shutil.rmtree, path, ignore_errors=True)
        os.makedirs(path, exist_ok=True)

        exporter.convert(dataset, save_dir=path, **kwargs)

    @staticmethod
    def validate(
        dataset: IDataset,
        task: Union[str, TaskType],
        *,
        env: Optional[Environment] = None,
        **kwargs,
    ) -> Dict:
        """
        Checks dataset annotations for correctness relatively to a task type.

        Args:
            dataset: The dataset to check
            task: Target task type - classification, detection etc.
            env: A plugin collection. If not set, the built-in plugins are used
            **kwargs: Parameters for the validator

        Returns: a dictionary with validation results
        """

        task = parse_str_enum_value(task, TaskType).name

        if env is None:
            env = Environment()

        validator: Validator = env.validators[task](**kwargs)
        return validator.validate(dataset)

    @staticmethod
    def aggregate(dataset: Dataset, from_subsets: Iterable[str], to_subset: str) -> Dataset:
        subset_names = set(dataset.subsets().keys())

        for subset in from_subsets:
            if subset not in subset_names:
                raise DatasetError(
                    f"{subset} is not found in the subset names ({subset_names}) in the dataset."
                )

        return HLOps.transform(
            dataset, "map_subsets", mapping={subset: to_subset for subset in from_subsets}
        )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import Any, Dict, List, Optional

import pyarrow as pa

from datumaro.components.dataset_base import DatasetItem
from datumaro.plugins.data_formats.datumaro_binary.mapper.annotation import AnnotationListMapper
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import DictMapper, Mapper

from .media import MediaMapper
from .utils import pa_batches_decoder


class DatasetItemMapper(Mapper):
    @staticmethod
    def forward(obj: DatasetItem, **options) -> Dict[str, Any]:
        return {
            "id": obj.id,
            "subset": obj.subset,
            "media": MediaMapper.forward(obj.media, **options.get("media", {})),
            "annotations": AnnotationListMapper.forward(obj.annotations),
            "attributes": DictMapper.forward(obj.attributes),
        }

    @staticmethod
    def backward(obj: Dict[str, Any]) -> DatasetItem:
        return DatasetItem(
            id=obj["id"],
            subset=obj["subset"],
            media=MediaMapper.backward(obj["media"]),
            annotations=AnnotationListMapper.backward(obj["annotations"])[0],
            attributes=DictMapper.backward(obj["attributes"])[0],
        )

    @staticmethod
    def backward_from_batches(
        batches: List[pa.lib.RecordBatch],
        parent: Optional[str] = None,
    ) -> List[DatasetItem]:
        ids = pa_batches_decoder(batches, f"{parent}.id" if parent else "id")
        subsets = pa_batches_decoder(batches, f"{parent}.subset" if parent else "subset")
        medias = MediaMapper.backward_from_batches(
            batches, f"{parent}.media" if parent else "media"
        )
        annotations_ = pa_batches_decoder(
            batches, f"{parent}.annotations" if parent else "annotations"
        )
        annotations_ = [
            AnnotationListMapper.backward(annotations)[0] for annotations in annotations_
        ]
        attributes_ = pa_batches_decoder(
            batches, f"{parent}.attributes" if parent else "attributes"
        )
        attributes_ = [DictMapper.backward(attributes)[0] for attributes in attributes_]

        items = []
        for id, subset, media, annotations, attributes in zip(
            ids, subsets, medias, annotations_, attributes_
        ):
            items.append(
                DatasetItem(
                    id=id,
                    subset=subset,
                    media=media,
                    annotations=annotations,
                    attributes=attributes,
                )
            )
        return items



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from datumaro.util.tf_util import import_tf

import_tf()  # prevent TF loading and potential interpreter crash

from itertools import groupby

from openvino.tools.accuracy_checker.adapters import create_adapter
from openvino.tools.accuracy_checker.data_readers import DataRepresentation
from openvino.tools.accuracy_checker.launcher import InputFeeder, create_launcher
from openvino.tools.accuracy_checker.postprocessor import PostprocessingExecutor
from openvino.tools.accuracy_checker.preprocessor import PreprocessingExecutor
from openvino.tools.accuracy_checker.utils import extract_image_representations

from datumaro.components.annotation import AnnotationType, LabelCategories

from .representation import import_predictions


class _FakeDataset:
    def __init__(self, metadata=None):
        self.metadata = metadata or {}


class GenericAcLauncher:
    @staticmethod
    def from_config(config):
        launcher_config = config["launcher"]
        launcher = create_launcher(launcher_config)

        dataset = _FakeDataset()
        adapter_config = config.get("adapter") or launcher_config.get("adapter")
        label_config = adapter_config.get("labels") if isinstance(adapter_config, dict) else None
        if label_config:
            assert isinstance(label_config, (list, dict))
            if isinstance(label_config, list):
                label_config = dict(enumerate(label_config))

            dataset.metadata = {
                "label_map": {int(key): label for key, label in label_config.items()}
            }
        adapter = create_adapter(adapter_config, launcher, dataset)

        preproc_config = config.get("preprocessing")
        preproc = None
        if preproc_config:
            preproc = PreprocessingExecutor(
                preproc_config,
                dataset_meta=dataset.metadata,
                input_shapes=launcher.inputs_info_for_meta(),
            )

        postproc_config = config.get("postprocessing")
        postproc = None
        if postproc_config:
            postproc = PostprocessingExecutor(
                postproc_config,
                dataset_meta=dataset.metadata,
            )

        return __class__(launcher, adapter=adapter, preproc=preproc, postproc=postproc)

    def __init__(self, launcher, adapter=None, preproc=None, postproc=None, input_feeder=None):
        self._launcher = launcher
        self._input_feeder = input_feeder or InputFeeder(
            launcher.config.get("inputs", []),
            launcher.inputs,
            launcher.fit_to_input,
            launcher.default_layout,
        )
        self._adapter = adapter
        self._preproc = preproc
        self._postproc = postproc

        self._categories = self._init_categories()

    def launch_raw(self, inputs):
        ids = range(len(inputs))
        inputs = [DataRepresentation(inp, identifier=id) for id, inp in zip(ids, inputs)]
        _, batch_meta = extract_image_representations(inputs)

        if self._preproc:
            inputs = self._preproc.process(inputs)

        inputs = self._input_feeder.fill_inputs(inputs)
        outputs = self._launcher.predict(inputs, batch_meta)

        if self._adapter:
            outputs = self._adapter.process(outputs, ids, batch_meta)

        if self._postproc:
            outputs = self._postproc.process(outputs)

        return outputs

    def launch(self, inputs):
        outputs = self.launch_raw(inputs)
        return [import_predictions(g) for _, g in groupby(outputs, key=lambda o: o.identifier)]

    def categories(self):
        return self._categories

    def _init_categories(self):
        if self._adapter is None or self._adapter.label_map is None:
            return None

        label_map = sorted(self._adapter.label_map.items(), key=lambda e: e[0])

        label_cat = LabelCategories()
        for _, label in label_map:
            label_cat.add(label)

        return {AnnotationType.label: label_cat}
